
	TITLE 'ASM48 MS-DOS Cross-Assembler'
;
; ASM48 is a cross-assembler for the INTEL MCS-48 and UPI-41 family
; of single-chip micro-controllers including the 8035, 8039, 8048
; and 8041.  Object code is output as an INTEL-HEX format ASCII file.
; The assembler will operate under MS-DOS 2.x compatible operating
; systems.
;
; Note: only 8086 compatible instructions are employed.
;

REL	EQU	1		; release number
REV	EQU	2		; revision level

DATE	 MACRO
	DB	'24-Sep-02'	; date
	 ENDM

NO	EQU	0		; false
YES	EQU	NOT NO		; true

NULL	EQU	00H		; null
HT	EQU	09H		; tab
LF	EQU	0AH		; line feed
FF	EQU	0CH		; form feed
CR	EQU	0DH		; carriage return
EOF	EQU	1AH		; end-of-file character


NASM	EQU	16		; buffers for asm input file
NHEX	EQU	4		; buffers for hex output file
NPRN	EQU	8		; buffers for prn outfile file

PWIDTH	EQU	130		; max list line length
;

	.8086

MAIN	SEGMENT
	ASSUME	CS:MAIN,DS:MAIN,SS:MAIN,ES:MAIN

	ORG	0000H

ORIG	=	$

FCB	EQU	ORIG+5CH	; default fcb
DBUF	EQU	ORIG+80H	; default disk buffer

	ORG	0100H		; start of program

START:	JMP	START1

	DB	8 DUP (0)

; Patch area for user selected defaults

INIFF	DB	0		; initial formfeed, 0=none		010BH
ENDFF	DB	1		; ending  formfeed, 0=none		010CH
PAGLEN	DB	60		; lines per page, 0=no page breaks	010DH
ASMTST	DB	0		; ASM style true/false test, 0=no	010EH
ASMCMD	DB	0		; ASM style command line, 0=no		010FH

;
START1:	MOV	SP,OFFSET STACKK

	MOV	BX,WORD PTR ORIG+6	; set top of memory
	MOV	BL,0
	MOV	WORD PTR MEMTOP,BX

	MOV	AH,30H			; get DOS version
	CALL	DOS
	MOV	BYTE PTR DOSVER,AL

; initialise program return code as 'unsuccessful'

	MOV	BYTE PTR PRC,1
;

START2:	CALL	CRLF			; display sign-on message
	MOV	BX,OFFSET MSG1
	CALL	PRINT
	MOV	BX,OFFSET MSG1A
	CALL	PRINT

; assign drives for source, hex and prn files

	MOV	AL,BYTE PTR FCB+1 	; source name?
	CMP	AL,' '
	JNZ	START3
	JMP	NAMERR

START3:	MOV	AH,19H			; get default drive
	CALL	DOS
	MOV	BYTE PTR DEF,AL

	MOV	BX,OFFSET FCB+8		; point to filetype of default fcb
	CALL	NXTDRV			; assign source drive

	CMP	AL,'Z'-'A'
	JNZ	START4

	MOV	AL,BYTE PTR DEF

START4:	MOV	BYTE PTR SRC,AL
	CALL	NXTDRV			; assign hex drive
	MOV	BYTE PTR HEX,AL
	CALL	NXTDRV			; assign prn drive
	MOV	BYTE PTR PRN,AL

	MOV	BX,OFFSET SRCFCB	; move filename to src fcb
	CALL	MOVNAM

	CALL	?PRN			; create prn file
	JZ	START5

	MOV	BX,OFFSET PRNFCB
	PUSH	BX
	PUSH	BX
	CALL	MOVNAM
	CALL	SELPRN
	POP	DX
	CALL	DELETF
	POP	DX
	CALL	MAKEF

START5:	MOV	AL,BYTE PTR HEX		; create hex file
	CMP	AL,'Z'-'A'
	JNZ	START6
	JMP	PASS1

START6:	MOV	BX,OFFSET HEXFCB
	PUSH	BX
	PUSH	BX
	CALL	MOVNAM
	CALL	SELHEX
	POP	DX
	CALL	DELETF
	POP	DX
	CALL	MAKEF
	JMP	PASS1			; begin assembly
;
; Select disk  (a)
;
SELDSK:	MOV	BX,OFFSET DEF		; already selected?
	CMP	AL,[BX]
	JNZ	SELDSK1
	RET				; yes, return

SELDSK1:MOV	[BX],AL
	MOV	DL,AL
	MOV	AH,0EH			; select disk fn
	JMP	DOS
;
; Move filename from fcb to (BX), no wildcards
;
MOVNAM:	MOV	DX,OFFSET FCB
	MOV	CH,9
MOVNAM1:XCHG	BX,DX
	MOV	AL,[BX]
	XCHG	BX,DX
	CMP	AL,'?'
	JNZ	MOVNAM2
	JMP	NAMERR

MOVNAM2:MOV	[BX],AL
	INC	BX
	INC	DX
	DEC	CH
	JNZ	MOVNAM1
	RET
;
; Open file
;
OPENF:	MOV	AH,0FH
	CALL	DOS
	INC	AL
	JZ	OPENF1
	RET

OPENF1:	MOV	BX,OFFSET MSG2		; no source file
	CALL	PRINT
	JMP	EXIT
;
; Close file
;
CLOSEF:	MOV	AH,10H
	CALL	DOS
	INC	AL
	JZ	CLOSEF1
	RET

CLOSEF1:MOV	BX,OFFSET MSG7		; cannot close files
	CALL	PRINT
	JMP	EXIT
;
; Delete file
;
DELETF:	MOV	AH,13H
	JMP	DOS
;
; Make new file
;
MAKEF:	MOV	AH,16H
	CALL	DOS
	INC	AL
	JZ	MAKEF1
	RET

MAKEF1:	MOV	BX,OFFSET MSG3		; no dir space
	CALL	PRINT
	JMP	EXIT
;
; Select source drive
;
SELSRC:	MOV	AL,BYTE PTR SRC
	JMP	SELDSK
;
; Select prn drive
;
SELPRN:	MOV	AL,BYTE PTR PRN
	JMP	SELDSK
;
; Select hex drive
;
SELHEX:	MOV	AL,BYTE PTR HEX
	JMP	SELDSK
;
; Test for prn file
;
?PRN:	MOV	AL,BYTE PTR PRN
	CMP	AL,'Z'-'A'
	JNZ	?PRN1
	RET

?PRN1:	CMP	AL,'X'-'A'
	RET
;
; Get drive from filetype
;
NXTDRV:	INC	BX
	MOV	AL,[BX]
	CMP	AL,' '
	JZ	NXTDRV1			; not specified

	SBB	AL,'A'
	RET

NXTDRV1:MOV	AL,BYTE PTR ASMCMD	; using ASM style command?
	OR	AL,AL
	MOV	AL,BYTE PTR DEF		; then default is current drive
	JZ	NXTDRV2
	RET

NXTDRV2:MOV	AL,'Z'-'A'		; else default is no output
	RET
;
; Open source file
;
OPENSRC:MOV	BX,NASM*128 		; set buffer pointer to end
	MOV	WORD PTR SRCPTR,BX	; to cause disk read
	XOR	AX,AX
	MOV	BYTE PTR SRCFCB+32,AL	; zero fcb extent
	MOV	WORD PTR SRCFCB+33,AX	; zero fcb record#
	MOV	WORD PTR SRCFCB+35,AX	; zero fcb
	MOV	BYTE PTR NUMHEX,AL
	CALL	SELSRC			; open
	MOV	DX,OFFSET SRCFCB
	JMP	OPENF
;
; Source filename error
;
NAMERR:	MOV	BX,OFFSET MSG4
	CALL	PRINT
	JMP	EXIT
;
; Input character (A) from source file
;
SRCBYT:	PUSH	CX
	PUSH	DX
	PUSH	BX
	MOV	BX,WORD PTR SRCPTR
	MOV	DX,NASM*128
	CMP	DX,BX
	JNZ	SRCBYT6
	CALL	SELSRC
	MOV	BX,0
	MOV	WORD PTR SRCPTR,BX
	MOV	CH,NASM
	MOV	BX,OFFSET SRCBUF
SRCBYT1:PUSH	CX
	PUSH	BX
	MOV	AH,14H			; read sequential
	MOV	DX,OFFSET SRCFCB
	CALL	DOS
	POP	BX
	POP	CX
	OR	AL,AL
	MOV	CL,128
	JZ	SRCBYT2			; 0 = ok
	CMP	AL,3			; 3 = partial record
	JNZ	SRCBYT4

SRCBYT2:MOV	DX,OFFSET DBUF
	MOV	CL,128
SRCBYT3:XCHG	BX,DX
	MOV	AL,[BX]
	XCHG	BX,DX
	MOV	[BX],AL
	INC	DX
	INC	BX
	DEC	CL
	JNZ	SRCBYT3
	DEC	CH
	JNZ	SRCBYT1
	JMP	SRCBYT6

SRCBYT4:CMP	AL,2			; 1 = eof, 2 = DTA segment wrap
	JZ	SRCBYT7

SRCBYT5:MOV	BYTE PTR [BX],EOF	; pad remainder sector with EOF char
	INC	BX
	DEC	CL
	JNZ	SRCBYT5

SRCBYT6:MOV	DX,OFFSET SRCBUF
	MOV	BX,WORD PTR SRCPTR
	PUSH	BX
	INC	BX
	MOV	WORD PTR SRCPTR,BX
	POP	BX
	ADD	BX,DX
	MOV	AL,[BX]
	POP	BX
	POP	DX
	POP	CX
	RET

SRCBYT7:MOV	BX,OFFSET MSG5		; source read error
	CALL	PRINT
	JMP	EXIT
;
; Output character (A) to destination file (prn or screen), CX saved
;
PRNBYT:	PUSH	CX
	MOV	CH,AL
	MOV	AL,BYTE PTR PRN
	CMP	AL,'Z'-41H		; output?
	JZ	PRNBYT2

	CMP	AL,'X'-41H		; screen?
	MOV	AL,CH
	JNZ	PRNBYT1

	CALL	CONBYT
	JMP	PRNBYT2

PRNBYT1:PUSH	CX
	CALL	PRNBYT4
	POP	CX
PRNBYT2:MOV	AL,CH
	CMP	AL,LF			; was character a linefeed?
	JNZ	PRNBYT3

	MOV	BX,OFFSET LINES		; then incr line counter
	INC	BYTE PTR [BX]
	CALL	HEADER			; output a header?
PRNBYT3:POP	CX
	RET

PRNBYT4:MOV	BX,WORD PTR PRNPTR
	XCHG	BX,DX
	MOV	BX,OFFSET PRNBUF
	ADD	BX,DX
	MOV	[BX],AL
	XCHG	BX,DX
	INC	BX
	MOV	WORD PTR PRNPTR,BX
	XCHG	BX,DX
	MOV	BX,NPRN*128
	CMP	DX,BX
	JZ	PRNBYT5
	RET

PRNBYT5:CALL	SELPRN
	MOV	BX,0
	MOV	WORD PTR PRNPTR,BX
	MOV	BX,OFFSET PRNBUF
	MOV	DX,OFFSET PRNFCB
	MOV	CH,NPRN			; number of buffers to write
;
; Write buffers to output file	BX=addr DX=fcb CH=#records
;
WRTBUF:	MOV	AL,[BX]
	CMP	AL,EOF
	JNZ	WRTBUF1
	RET

WRTBUF1:PUSH	CX
	PUSH	DX
	MOV	CL,128
	MOV	DX,OFFSET DBUF
WRTBUF2:MOV	AL,[BX]
	XCHG	BX,DX
	MOV	[BX],AL
	XCHG	BX,DX
	INC	BX
	INC	DX
	DEC	CL
	JNZ	WRTBUF2

	POP	DX
	PUSH	DX
	PUSH	BX
	MOV	AH,15H			; write sequential
	CALL	DOS
	POP	BX
	POP	DX
	POP	CX
	OR	AL,AL
	JZ	WRTBUF3
	JMP	WRTERR

WRTBUF3:DEC	CH
	JNZ	WRTBUF
	RET

WRTERR:	MOV	BX,OFFSET MSG6		; output file write error
	CALL	PRINT
	JMP	DONE5
;
; Output byte to hex file
;
HEXBYT:	PUSH	CX
	PUSH	DX
	PUSH	BX
	CALL	HEXBYT1
	POP	BX
	POP	DX
	POP	CX
	RET

HEXBYT1:MOV	BX,WORD PTR HEXPTR
	XCHG	BX,DX
	MOV	BX,OFFSET HEXBUF
	ADD	BX,DX
	MOV	[BX],AL
	XCHG	BX,DX
	INC	BX
	MOV	WORD PTR HEXPTR,BX
	XCHG	BX,DX
	MOV	BX,NHEX*128
	CMP	DX,BX
	JZ	HEXBYT2
	RET

HEXBYT2:CALL	SELHEX
	MOV	BX,0
	MOV	WORD PTR HEXPTR,BX
	MOV	BX,OFFSET HEXBUF
	MOV	DX,OFFSET HEXFCB
	MOV	CH,NHEX
	JMP	WRTBUF
;
; Output CR/LF to console
;
CRLF:	MOV	AL,CR
	CALL	CONBYT
	MOV	AL,LF
;
; Output character in A to console,  CX saved
;
CONBYT:	PUSH	CX
	MOV	AH,2			; console output
	MOV	DL,AL
	CALL	DOS
	POP	CX
	RET
;
; Output string (BX) to console
;
PRINT:	MOV	AL,[BX]
	OR	AL,AL
	JNZ	PRINT1
	RET

PRINT1:	PUSH	BX
	CALL	CONBYT
	POP	BX
	INC	BX
	JMP	PRINT
;
; Output line of source to prn file (and screen if error flagged)
;
WRTLN:	CALL	?ERROR			; any errors on this line?
	JZ	WRTLN1

	MOV	BX,WORD PTR ERRCNT	; yes, bump count
	INC	BX
	MOV	WORD PTR ERRCNT,BX

	MOV	BX,OFFSET MSG11A 	; update error msg field
	CALL	INCNUM

WRTLN1:	MOV	BX,OFFSET LINNO+4 	; bump line number
	CALL	INCNUM
	MOV	BX,OFFSET LINNO		; move it into place
	MOV	DX,OFFSET NUMFLD
	MOV	CH,5
	CALL	MOVE

	MOV	BX,OFFSET LINELEN
	MOV	CH,[BX]			; length
	INC	BX			; start of buffer
	INC	CH
WRTLN2:	DEC	CH
	JZ	WRTLN3

	MOV	AL,[BX]
	PUSH	BX
	CALL	WRTLN4
	POP	BX
	INC	BX
	JMP	WRTLN2

WRTLN3:	MOV	AL,CR			; cr/lf
	CALL	WRTLN4
	MOV	AL,LF
	CALL	WRTLN4
	JMP	CLRLIN

WRTLN4:	MOV	CL,AL
	CALL	PRNBYT			; send char to file
	CALL	?ERROR			; if error flagged
	JNZ	WRTLN5
	RET

WRTLN5:	MOV	AL,BYTE PTR PRN
	CMP	AL,'X'-41H		; and screen output off
	JNZ	WRTLN6
	RET

WRTLN6:	MOV	AL,CL
	JMP	CONBYT			; then output to console
;
; Clear output line buffer
;
CLRLIN:	MOV	BX,OFFSET LINELEN
	MOV	BYTE PTR [BX],0		; zero length
	MOV	AL,PWIDTH		; max length
CLRLIN1:INC	BX			; line address
	MOV	BYTE PTR [BX],' '	; blank with spaces
	DEC	AL
	JNZ	CLRLIN1
	RET
;
; Increment ascii number in string, BX = addr least significant digit
;
INCNUM:	MOV	AL,[BX]
	OR	AL,'0'			; convert space to '0'
	INC	AL
	MOV	[BX],AL
	CMP	AL,'9'+1
	JNC	INCNUM1
	RET

INCNUM1:MOV	BYTE PTR [BX],'0'
	DEC	BX
	JMP	INCNUM
;
; Output page header
;
HEADER:	MOV	AL,BYTE PTR PASS	; skip on first pass
	OR	AL,AL
	JNZ	HEADER1
	RET

HEADER1:MOV	AL,BYTE PTR PAGLEN	; no header if lines/page = 0
	OR	AL,AL
	JNZ	HEADER2
	RET

HEADER2:MOV	BX,OFFSET LINES		; # lines output so far
	CMP	AL,[BX]			; limit reached?
	JZ	HEADER3
	RET

HEADER3:MOV	BYTE PTR [BX],0		; reset line counter
	MOV	BX,OFFSET MSG12A	; adjust page number
	CALL	INCNUM

	MOV	BX,OFFSET INIFF		; send initial formfeed
	MOV	AL,[BX]			; depending on flag
	OR	AL,AL
	MOV	AL,1			; set flag for subsequent output
	MOV	[BX],AL
	MOV	AL,FF
	JZ	HEADER4
	CALL	PRNBYT

HEADER4:MOV	BX,OFFSET MSG1		; first part of string
	CALL	WRTSTR
	MOV	BX,OFFSET MSG12		; second part of string
	CALL	WRTSTR
	CALL	WRTCR
	JMP	WRTCR
;
; Write CR/LF to prn file
;
WRTCR:	MOV	AL,CR
	CALL	PRNBYT
	MOV	AL,LF
	JMP	PRNBYT
;
; Write string (BX) to prn file
;
WRTSTR:	MOV	AL,[BX]
	OR	AL,AL
	JNZ	WRTSTR1
	RET

WRTSTR1:PUSH	BX
	CALL	PRNBYT
	POP	BX
	INC	BX
	JMP	WRTSTR
;
; Flush output buffers, close files and exit to DOS
;
DONE:	CALL	?PRN			; if prn file
	JZ	DONE2

	MOV	AL,BYTE PTR ENDFF	; and ending FF flag
	OR	AL,AL
	MOV	AL,FF			; send a terminating form feed
	JZ	DONE1
	CALL	PRNBYT
DONE1:	MOV	BX,WORD PTR PRNPTR 	; pad remainder with EOF char
	MOV	AL,BL
	OR	AL,BH
	JZ	DONE2
	MOV	AL,EOF
	CALL	PRNBYT
	JMP	DONE1

DONE2:	MOV	AL,BYTE PTR HEX		; output current hex line
	CMP	AL,'Z'-'A'
	JZ	DONE5
	MOV	AL,BYTE PTR NUMHEX
	OR	AL,AL
	JZ	DONE3
	CALL	OHEXLN
DONE3:	MOV	BX,OFFSET HEXREC	; output hex 'end-of-record'
	INC	BYTE PTR [BX]
	MOV	BX,WORD PTR EXEC
	MOV	WORD PTR HEXADDR,BX
	CALL	OHEXLN
DONE4:	MOV	BX,WORD PTR HEXPTR	; pad remainder with EOF char
	MOV	AL,BL
	OR	AL,BH
	JZ	DONE5
	MOV	AL,EOF
	CALL	HEXBYT
	JMP	DONE4

DONE5:	CALL	?PRN			; close prn file
	JZ	DONE6
	CALL	SELPRN
	MOV	DX,OFFSET PRNFCB
	CALL	CLOSEF

DONE6:	MOV	AL,BYTE PTR HEX		; close hex file
	CMP	AL,'Z'-'A'
	JZ	DONE7
	CALL	SELHEX
	MOV	DX,OFFSET HEXFCB
	CALL	CLOSEF

DONE7:	CALL	SELSRC			; close source file
	MOV	DX,OFFSET SRCFCB
	CALL	CLOSEF

	CALL	CRLF			; end of assembly
	MOV	BX,OFFSET MSG8
	CALL	PRINT
;
; Exit to DOS
;
EXIT:	MOV	AL,BYTE PTR PRC		; get return code
	MOV	AH,4CH			; and quit
	INT	21H

;
; Convert byte to hex ascii and output to hex file
;
WRTHEX:	PUSH	CX
	MOV	CH,AL
	MOV	AL,BYTE PTR HEX
	CMP	AL,'Z'-41H
	MOV	AL,CH
	JZ	WRTHEX5

	PUSH	DX
	LAHF
	PUSH	AX
	MOV	BX,OFFSET NUMHEX
	MOV	AL,[BX]
	OR	AL,AL
	JZ	WRTHEX3

	CMP	AL,16			; max hex bytes/line
	JC	WRTHEX1

	CALL	OHEXLN
	JMP	WRTHEX3

WRTHEX1:MOV	BX,WORD PTR CURPC
	XCHG	BX,DX
	MOV	BX,WORD PTR HEXADDR
	MOV	CL,AL
	MOV	CH,0
	ADD	BX,CX
	MOV	AL,DL
	CMP	AL,BL
	JNZ	WRTHEX2

	MOV	AL,DH
	CMP	AL,BH
	JZ	WRTHEX4

WRTHEX2:CALL	OHEXLN
WRTHEX3:MOV	BX,WORD PTR CURPC
	MOV	WORD PTR HEXADDR,BX
WRTHEX4:MOV	BX,OFFSET NUMHEX
	MOV	DL,[BX]
	INC	BYTE PTR [BX]
	MOV	DH,0
	MOV	BX,OFFSET HEXLIST
	ADD	BX,DX
	POP	AX
	SAHF
	MOV	[BX],AL
	POP	DX
WRTHEX5:POP	CX
	RET
;
; Convert byte (A) to hex ascii and output
;
TOHEX:	LAHF
	PUSH	AX
	ROR	AL,1
	ROR	AL,1
	ROR	AL,1
	ROR	AL,1
	AND	AL,00001111B
	CALL	TOHEX1
	POP	AX
	SAHF
	LAHF
	PUSH	AX
	AND	AL,00001111B
	CALL	TOHEX1
	POP	AX
	SAHF
	ADD	AL,DH
	MOV	DH,AL
	RET

TOHEX1:	ADD	AL,90H
	DAA
	ADC	AL,'@'
	DAA
	JMP	HEXBYT
;
; Output an 'INTEL HEX' line including address and checksum
;
OHEXLN:	MOV	AL,':'			; specify start of line
	CALL	HEXBYT
	MOV	BX,OFFSET NUMHEX	; number of bytes
	MOV	DL,[BX]
	XOR	AL,AL			; clear checksum
	MOV	DH,AL
	MOV	[BX],AL
	MOV	BX,WORD PTR HEXADDR
	MOV	AL,DL			; # bytes in line
	CALL	TOHEX
	MOV	AL,BH			; address
	CALL	TOHEX
	MOV	AL,BL
	CALL	TOHEX
	MOV	AL,BYTE PTR HEXREC	; record type  0=data, 1=end
	CALL	TOHEX
	MOV	AL,DL
	OR	AL,AL
	JZ	OHEXLN2

	MOV	BX,OFFSET HEXLIST	; output numhex bytes
OHEXLN1:MOV	AL,[BX]
	INC	BX
	CALL	TOHEX
	DEC	DL
	JNZ	OHEXLN1

OHEXLN2:XOR	AL,AL			; output checksum
	SUB	AL,DH
	CALL	TOHEX
	MOV	AL,CR			; output cr,lf
	CALL	HEXBYT
	MOV	AL,LF
	JMP	HEXBYT
;
; Parse a string from source file and place into buffer.  String
; type is determined (1-4).  If a number, then it is evaluated to
; 16 bits and placed in strval.
;
ALPHA	EQU	1		; alphabetic
NUMER	EQU	2		; numeric
QUOTE	EQU	3		; quote
OTHER	EQU	4		; other

PARSE:	XOR	AL,AL			; clear string type
	MOV	BYTE PTR STRTYP,AL
	CALL	CLRSTR		; zero string length, reset numeric base

PARSE1:	MOV	AL,BYTE PTR CHAR
	CMP	AL,HT			; tab?
	JZ	PARSE4
	CMP	AL,';'			; comment?
	JZ	PARSE2
	CMP	AL,'*'			; full line comment?
	JNZ	PARSE3
	MOV	AL,BYTE PTR PRVCHAR ; only if in first column, ie. previous
	CMP	AL,LF			; character must be a linefeed
	JNZ	PARSE3
PARSE2:	CALL	GETCHR			; get next char
	CALL	?EOSTMT			; until end of statement
	JZ	PARSE5
	JMP	PARSE2

PARSE3:	OR	AL,00100000B		; space or null
	CMP	AL,00100000B
	JNZ	PARSE5
PARSE4:	CALL	GETCHR			; loop
	JMP	PARSE1

; determine string type by examining first character

PARSE5:	CALL	?ALPHA			; alpha?
	JZ	PARSE6
	MOV	AL,ALPHA
	JMP	PARSE11

PARSE6:	CALL	?NUM			; numeric?
	JZ	PARSE7
	MOV	AL,NUMER
	JMP	PARSE11

PARSE7:	MOV	AL,BYTE PTR CHAR 	; quote?
	CMP	AL,''''
	JNZ	PARSE8
	XOR	AL,AL
	MOV	BYTE PTR CHAR,AL 	; yes set character to be ignored
	MOV	AL,QUOTE
	JMP	PARSE11

PARSE8:	CMP	AL,LF			; line feed?
	JNZ	PARSE10
	MOV	AL,BYTE PTR PASS	; then output line (if pass2)
	OR	AL,AL
	JZ	PARSE9
	CALL	WRTLN
PARSE9:	MOV	BX,OFFSET ERRFLD 	; clear error field
	MOV	BYTE PTR [BX],' '
	MOV	AL,SRCFLD-LINEBUF 	; reset line position
	MOV	BYTE PTR LINELEN,AL

PARSE10:MOV	AL,OTHER		; not alpha, number or quote
PARSE11:MOV	BYTE PTR STRTYP,AL

; build remainder of string

PARSE12:MOV	AL,BYTE PTR CHAR	; get character
	MOV	BYTE PTR PRVCHAR,AL	; save it
	OR	AL,AL
	JZ	PARSE13
	CALL	ADDSTR			; add to string unless 'ignore'
PARSE13:CALL	GETCHR			; get next character
	MOV	AL,BYTE PTR STRTYP	; if string is of invalid type
	CMP	AL,OTHER		; then stop building
	JNZ	PARSE14
	RET

PARSE14:CMP	AL,QUOTE		; convert character to uppercase
	JZ	PARSE15
	CALL	UPPER			; unless within quotes
PARSE15:MOV	BX,OFFSET CHAR
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,ALPHA		; alpha?
	JNZ	PARSE16
	CALL	?DOLLAR			; ignore '$'
	JZ	PARSE12
	CALL	?ALPHNM
	JNZ	PARSE12
	RET

PARSE16:CMP	AL,NUMER		; number
	JZ	PARSE17
	JMP	PARSE33

PARSE17:CALL	?DOLLAR			; ignore '$'
	JZ	PARSE12
	CALL	?HEX
	JNZ	PARSE12

	MOV	AL,BYTE PTR CHAR 	; is number octal?
	CMP	AL,'O'
	JZ	PARSE18
	CMP	AL,'Q'
	JNZ	PARSE19
PARSE18:MOV	AL,8
	JMP	PARSE20

PARSE19:CMP	AL,'H'			; is number hex?
	JNZ	PARSE21
	MOV	AL,16

PARSE20:MOV	BYTE PTR NUMBASE,AL	; set number base
	XOR	AL,AL
	MOV	BYTE PTR CHAR,AL
	JMP	PARSE25

PARSE21:MOV	AL,BYTE PTR PRVCHAR
	CMP	AL,'B'			; binary
	JNZ	PARSE22
	MOV	AL,2
	JMP	PARSE23

PARSE22:CMP	AL,'D'			; decimal
	MOV	AL,10
	JNZ	PARSE24

PARSE23:MOV	BX,OFFSET STRLEN	; strip base identifier from string
	DEC	BYTE PTR [BX]
PARSE24:MOV	NUMBASE,AL		; set base
PARSE25:MOV	BX,0			; evaluate to a binary value
	MOV	WORD PTR STRVAL,BX
	MOV	BX,OFFSET STRLEN
	MOV	CL,[BX]
	INC	BX
PARSE26:MOV	AL,[BX]			; get ascii char from str buffer
	INC	BX
	CMP	AL,'A'			; and convert to a binary value
	JNC	PARSE27
	SUB	AL,'0'
	JMP	PARSE28

PARSE27:SUB	AL,'7'
PARSE28:PUSH	BX			; save position in str buffer
	PUSH	CX			; save str length
	MOV	CL,AL			; check value lies within current
	MOV	BX,OFFSET NUMBASE	; number base
	CMP	AL,[BX]
	JC	PARSE29
	CALL	VALUERR
PARSE29:MOV	CH,0
	MOV	AL,[BX]
	MOV	BX,WORD PTR STRVAL
	XCHG	BX,DX
	MOV	BX,0
PARSE30:OR	AL,AL
	JZ	PARSE32
	RCR	AL,1
	JNC	PARSE31
	ADD	BX,DX
PARSE31:XCHG	BX,DX
	ADD	BX,BX
	XCHG	BX,DX
	JMP	PARSE30

PARSE32:ADD	BX,CX
	MOV	WORD PTR STRVAL,BX
	POP	CX
	POP	BX
	DEC	CL
	JNZ	PARSE26
	RET

PARSE33:MOV	AL,BYTE PTR CHAR
	CMP	AL,CR
	JNZ	PARSE34
	JMP	OVFLERR

PARSE34:CMP	AL,''''
	JZ	PARSE35
	JMP	PARSE12

PARSE35:CALL	GETCHR
	CMP	AL,''''
	JZ	PARSE36
	RET

PARSE36:JMP	PARSE12
;
; Convert character to uppercase (except if within quotes)
;
TOUPPER:LAHF
	PUSH	AX
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,QUOTE
	JZ	TOUPPER1
	CALL	UPPER
TOUPPER1:
	POP	AX
	SAHF
	RET
;
; Get next character from source, place in line buffer, convert
; to uppercase
;
GETCHR:	CALL	SRCBYT			; get char from input file
	MOV	BYTE PTR CHAR,AL	; save it
	CMP	AL,CR			; don't include cr,lf in line
	JZ	GETCHR1

	CMP	AL,LF
	JZ	GETCHR1

	MOV	AL,BYTE PTR LINELEN	; buffer full?
	CMP	AL,PWIDTH
	JNC	GETCHR1

	MOV	DL,AL			; no, add char to line
	MOV	DH,0
	INC	AL
	MOV	BX,OFFSET LINELEN
	MOV	[BX],AL			; update length
	INC	BX
	ADD	BX,DX
	MOV	AL,BYTE PTR CHAR
	MOV	[BX],AL
GETCHR1:JMP	TOUPPER			; make uppercase
;
; Initialise for next line
;
INITLN:	CALL	CLRSTR			; clear string buffer
	CALL	CLRLIN			; blank line buffer
	XOR	AL,AL			; zero current character
	MOV	BYTE PTR CHAR,AL
	MOV	AL,LF			; set end-of-line
	MOV	BYTE PTR PRVCHAR,AL
	MOV	AL,SRCFLD-LINEBUF 	; set source list offset
	MOV	BYTE PTR LINELEN,AL
	RET
;
; Clear string, number base
;
CLRSTR:	XOR	AL,AL
	MOV	BYTE PTR STRLEN,AL
	MOV	BYTE PTR NUMBASE,AL
	RET
;
; Add char to string
;
ADDSTR:	MOV	BX,OFFSET STRLEN
	MOV	AL,[BX]
	CMP	AL,64			; max length of a parsed string
	JC	ADDSTR1

	MOV	BYTE PTR [BX],0		; overflow
	CALL	OVFLERR
ADDSTR1:MOV	DL,[BX]
	MOV	DH,0
	INC	BYTE PTR [BX]		; increment count
	INC	BX			; start of string buffer
	ADD	BX,DX
	MOV	AL,BYTE PTR CHAR
	MOV	[BX],AL			; and store character
	RET
;
; Insert null if '$' characters found (BX)
;
?DOLLAR:MOV	AL,[BX]
	CMP	AL,'$'
	JNZ	?DOLLAR1
	XOR	AL,AL			; substitute 0
	MOV	[BX],AL
?DOLLAR1:
	RET
;
; Test numeric '0' to '9'
;
?NUM:	MOV	AL,BYTE PTR CHAR
	SUB	AL,'0'
	CMP	AL,10
	RCL	AL,1
	AND	AL,1
	RET
;
; Test hex
;
?HEX:	CALL	?NUM
	JZ	?HEX1
	RET

?HEX1:	MOV	AL,BYTE PTR CHAR
	SUB	AL,'A'
	CMP	AL,6
	RCL	AL,1
	AND	AL,1
	RET
;
; Test alpha 'A' to 'Z' (including '?' and '@')
;
?ALPHA:	MOV	AL,BYTE PTR CHAR
	SUB	AL,'?'
	CMP	AL,28
	RCL	AL,1
	AND	AL,1
	RET
;
; Test alpha-numeric
;
?ALPHNM:CALL	?ALPHA
	JZ	?ALPHNM1
	RET

?ALPHNM1:
	JMP	?NUM
;
; Convert (char) to upper case
;
UPPER:	MOV	AL,BYTE PTR CHAR
	CMP	AL,'a'
	JNC	UPPER1
	RET

UPPER1:	CMP	AL,'z'+1
	JC	UPPER2
	RET

UPPER2:	AND	AL,5FH
	MOV	BYTE PTR CHAR,AL
	RET
;
; Test for end of statement character (cr, eof, !, null)
;
?EOSTMT:CMP	AL,CR
	JZ	?EOSTMT1
	CMP	AL,'!'
	JZ	?EOSTMT1
	CMP	AL,EOF
	JZ	?EOSTMT1
	CMP	AL,NULL
?EOSTMT1:
	RET
;
; Symbol table
;
; The symbol table is organised as a linkled list.  Each new symbol
; added to the table contains a link to the previously entered
; symbol.  To minimize long searches, the symbol table is organised
; as 128 separate lists.  To determine into which list a symbol will
; be added, the symbol name is hashed to a value 0 to 127.  A link
; table (lnktbl) comprising 128 pointers gives the addresses of the
; last symbol added to each of the lists.
;
; Each symbol in the symbol table is comprised of the following
;
;	2  bytes	pointer (link) to previous symbol
;	1		lower nibble - string length
;			upper nibble - flags
;			  bit 0 - value defined
;			      1 -
;			      2 - value defined by SET
;			      3 - not used
;	n		name (up to 16 characters)
;	2		value of symbol (initialised to zero)

;
; Clear all symbols table
;
CLRSYM:	MOV	BX,OFFSET LNKTBL ; clear link table
	MOV	CH,128
	XOR	AL,AL
CLRSYM1:MOV	[BX],AL
	INC	BX
	MOV	[BX],AL
	INC	BX
	DEC	CH
	JNZ	CLRSYM1

	MOV	BX,0			; clear pointer
	MOV	WORD PTR SYMPTR,BX
	RET
;
; Hash string (any length) to a 7 bit value
;
HASH:	MOV	BX,OFFSET STRLEN
	MOV	CH,[BX]
	XOR	AL,AL
HASH1:	INC	BX
	ADD	AL,[BX]
	DEC	CH
	JNZ	HASH1

	AND	AL,01111111B
	MOV	BYTE PTR HASHVAL,AL
	RET
;
; Get upper nibble
;
	MOV	CH,AL
	MOV	BX,WORD PTR SYMPTR
	INC	BX
	INC	BX
	MOV	AL,[BX]
	AND	AL,11110000B
	OR	AL,CH
	MOV	[BX],AL
	RET
;
; Get # chars in symbol name
;
SYMLEN:	MOV	BX,WORD PTR SYMPTR
	INC	BX			; move to length field
	INC	BX
	MOV	AL,[BX]
	AND	AL,00001111B
	INC	AL
	RET
;
; Test symbol pointer, Z if found
;
?SYM:	MOV	BX,WORD PTR SYMPTR
	MOV	AL,BL
	OR	AL,BH
	RET
;
; Search symbol table for string, if found (symptr) leave
; address otherwise zero
;
FNDSYM:	CALL	HASH			; hash expression to 7 bits
	MOV	BX,OFFSET STRLEN
	MOV	AL,[BX]
	CMP	AL,16+1
	JC	FNDSYM1

	MOV	BYTE PTR [BX],16	; limit to 16 characters
FNDSYM1:MOV	BX,OFFSET HASHVAL	; get str's hashed value (0-127)
	MOV	DL,[BX]
	MOV	DH,0
	MOV	BX,OFFSET LNKTBL	; use as position in lnktbl
	ADD	BX,DX
	ADD	BX,DX
	MOV	DL,[BX]			; get 2 bytes
	INC	BX
	MOV	BH,[BX]
	MOV	BL,DL
FNDSYM2:MOV	WORD PTR SYMPTR,BX	; store it
	CALL	?SYM			; return if zero
	JNZ	FNDSYM3
	RET

FNDSYM3:CALL	SYMLEN			; get length of symbol + 1
	MOV	BX,OFFSET STRLEN	; compare symbol length with BX
	CMP	AL,[BX]
	JNZ	FNDSYM5			; no match, skip

	MOV	CH,AL			; compare symbol string with BX
	INC	BX
	XCHG	BX,DX
	MOV	BX,WORD PTR SYMPTR
	INC	BX
	INC	BX
	INC	BX
FNDSYM4:XCHG	BX,DX
	MOV	AL,[BX]
	XCHG	BX,DX
	CMP	AL,[BX]
	JNZ	FNDSYM5			; nomatch, skip
	INC	DX
	INC	BX
	DEC	CH
	JNZ	FNDSYM4
	RET

FNDSYM5:MOV	BX,WORD PTR SYMPTR
	MOV	DL,[BX]
	INC	BX
	MOV	DH,[BX]
	XCHG	BX,DX
	JMP	FNDSYM2
;
; Add symbol into table
;
ADDSYM:	MOV	BX,OFFSET STRLEN
	MOV	DL,[BX]
	MOV	DH,0
	MOV	BX,WORD PTR SYMTOP 	; insert new symbol at top of table
	MOV	WORD PTR SYMPTR,BX
	ADD	BX,DX			; add length of symbol
	MOV	DX,5
	ADD	BX,DX			; add 5 more bytes
	XCHG	BX,DX

	MOV	BX,WORD PTR MEMTOP 	; check for sufficient space
	MOV	AL,DL
	SUB	AL,BL
	MOV	AL,DH
	SBB	AL,BH
	XCHG	BX,DX
	JC	ADDSYM1
	JMP	SYMOVFL			; no room, exit with error message

ADDSYM1:MOV	WORD PTR SYMTOP,BX	; update top of symbol table
	MOV	BX,WORD PTR SYMPTR	; pointer to new symbol
	XCHG	BX,DX

	MOV	BX,OFFSET HASHVAL	; hashed value of symbol determines
	MOV	CL,[BX]			; offset into link table
	MOV	CH,0
	MOV	BX,OFFSET LNKTBL
	ADD	BX,CX
	ADD	BX,CX

	MOV	CL,[BX]			; point to appropriate list
	INC	BX
	MOV	CH,[BX]

	MOV	[BX],DH			; update pointer in link table to
	DEC	BX			; point to new symbol
	MOV	[BX],DL

	XCHG	BX,DX			; insert pointer to previous symbol
	MOV	[BX],CL			; into link field of new symbol
	INC	BX
	MOV	[BX],CH

	MOV	DX,OFFSET STRLEN	; insert length byte
	XCHG	BX,DX
	MOV	AL,[BX]
	XCHG	BX,DX
	CMP	AL,16+1			; 16 chars maximum
	JC	ADDSYM2

	MOV	AL,16
ADDSYM2:MOV	CH,AL
	DEC	AL
	INC	BX
	MOV	[BX],AL
ADDSYM3:INC	BX			; copy symbol name
	INC	DX
	XCHG	BX,DX
	MOV	AL,[BX]
	XCHG	BX,DX
	MOV	[BX],AL
	DEC	CH
	JNZ	ADDSYM3

	XOR	AL,AL			; initialise symbol value to zero
	INC	BX
	MOV	[BX],AL
	INC	BX
	MOV	[BX],AL
	RET

SYMOVFL:MOV	BX,OFFSET MSG9
	CALL	PRINT
	JMP	DONE
;
; Store high nibble (A)
;
PUTNBL:	RCL	AL,1
	RCL	AL,1
	RCL	AL,1
	RCL	AL,1
	AND	AL,11110000B
	MOV	CH,AL
	MOV	BX,WORD PTR SYMPTR
	INC	BX
	INC	BX
	MOV	AL,[BX]
	AND	AL,00001111B
	OR	AL,CH
	MOV	[BX],AL
	RET
;
; Get high nibble (A)
;
GETNBL:	MOV	BX,WORD PTR SYMPTR
	INC	BX
	INC	BX
	MOV	AL,[BX]
	RCR	AL,1
	RCR	AL,1
	RCR	AL,1
	RCR	AL,1
	AND	AL,00001111B
	RET
;
; Point to symbol's value field (BX)
;
SVADR:	CALL	SYMLEN
	MOV	BX,WORD PTR SYMPTR
	MOV	DL,AL
	MOV	DH,0
	ADD	BX,DX
	INC	BX
	INC	BX
	INC	BX
	RET
;
; Store value (BX) of symbol
;
PUTSVAL:PUSH	BX
	CALL	SVADR
	POP	DX
	MOV	[BX],DL
	INC	BX
	MOV	[BX],DH
	RET
;
; Retrieve value of symbol (BX)
;
GETSVAL:CALL	SVADR
	MOV	DL,[BX]
	INC	BX
	MOV	DH,[BX]
	XCHG	BX,DX
	RET
;
; Test for mnemonic - Z if found, A=mode
;
?MNEM:	MOV	AL,BYTE PTR STRLEN	; test # chars in string (max 5)
	MOV	CL,AL
	DEC	AL
	MOV	DL,AL
	MOV	DH,0
	PUSH	DX
	CMP	AL,5
	JNC	?MNEM1

	MOV	BX,OFFSET NUMOPCS
	ADD	BX,DX
	MOV	CH,[BX]
	MOV	BX,OFFSET GRPTR		; pointer to mnemonic group
	ADD	BX,DX
	ADD	BX,DX
	MOV	DH,[BX]
	INC	BX
	MOV	BH,[BX]
	MOV	BL,DH
	MOV	DH,CL
	CALL	MATCH
	JNZ	?MNEM1
;
	POP	DX
	MOV	BX,OFFSET VALOPCS
	ADD	BX,DX
	ADD	BX,DX
	MOV	DL,[BX]
	INC	BX
	MOV	DH,[BX]
	MOV	BL,AL
	MOV	BH,0
	ADD	BX,BX
	ADD	BX,DX
	MOV	AL,[BX]
	INC	BX
	MOV	CH,[BX]
	CMP	AL,AL		; set Z flag
	RET

?MNEM1:	POP	DX
	XOR	AL,AL
	INC	AL
	RET
;
; Scan a mnemonic group (BX=adr, CH=#items, DH=item length) for a match
; with the string buffer
;
MATCH:	MOV	DL,0FFH
	INC	CH
	MOV	CL,0
MATCH1:	XOR	AL,AL
	MOV	AL,CH
	ADD	AL,CL
	RCR	AL,1
	CMP	AL,DL
	JZ	MATCH6
	MOV	DL,AL
	PUSH	BX
	PUSH	DX
	PUSH	CX
	PUSH	BX
	MOV	CH,DH
	MOV	CL,CH
	MOV	DH,0
	MOV	BX,0
MATCH2:	ADD	BX,DX
	DEC	CH
	JNZ	MATCH2
	POP	DX
	ADD	BX,DX
	MOV	DX,OFFSET STRING
MATCH3:	XCHG	BX,DX
	MOV	AL,[BX]
	XCHG	BX,DX
	CMP	AL,[BX]
	JNZ	MATCH4
	INC	DX
	INC	BX
	DEC	CL
	JNZ	MATCH3
	POP	CX
	POP	DX
	POP	BX
	MOV	AL,DL
	RET

MATCH4:	POP	CX
	POP	DX
	POP	BX
	JC	MATCH5
	MOV	CL,DL
	JMP	MATCH1

MATCH5:	MOV	CH,DL
	JMP	MATCH1

MATCH6:	XOR	AL,AL
	INC	AL
	RET

; Mnemonic tables

GRPTR:				; pointer to each mnemonic group
	DW	GRP1		; 1 character mnemonic
	DW	GRP2		; 2
	DW	GRP3		; 3
	DW	GRP4		; 4
	DW	GRP5		; 5

	DW	GRP11		; mode table

NUMOPCS	DB	GRP2-GRP1	; number of mnemonics in each group
	DB	(GRP3-GRP2)/2
	DB	(GRP4-GRP3)/3
	DB	(GRP5-GRP4)/4
	DB	(GRP11-GRP5)/5

VALOPCS:			; pointer to mode, opcode
	DW	GRP11
	DW	GRP12
	DW	GRP13
	DW	GRP14
	DW	GRP15

; Mnemonic groups sorted by length

GRP1	DB	CR,'#()*+,-/ACIT'

GRP2	DB	'@A','DA','DB','DS','DW','EN','F0','F1','IF'
	DB	'IN','JC','JZ','OR','P1','P2','P4','P5','P6'
	DB	'P7','R0','R1','R2','R3','R4','R5','R6','R7'
	DB	'RL','RR'

GRP3	DB	'@R0','@R1','ADD','AND','ANL','BUS','CLK','CLR'
	DB	'CNT','CPL','DBB','DEC','DIS','DMA','END','EQU'
	DB	'IDL','INC','INS','JB0','JB1','JB2','JB3','JB4'
	DB	'JB5','JB6','JB7','JF0','JF1','JMP','JNC','JNI'
	DB	'JNZ','JT0','JT1','JTF','MB0','MB1','MOD','MOV'
	DB	'NOP','NOT','ORG','ORL','OUT','PSW','RB0','RB1'
	DB	'RET','RLC','RRC','SEL','SET','SHL','SHR','STS'
	DB	'XCH','XOR','XRL'

GRP4	DB	'ADDC','ANLD','CALL','DJNZ','ELSE','ENDM','ENT0'
	DB	'JMPP','JNT0','JNT1','JOBF','MOVD','MOVP','MOVX'
	DB	'ORLD','OUTL','PAGE','RETR','STOP','STRT','SWAP'
	DB	'TCNT','XCHD'

GRP5	DB	'ENDIF','FLAGS','JNIBF','MACRO','MOVP3','TCNTI'
	DB	'TITLE'

; Reference mnemonics by mode, value

GRP11	DB	15,	10	; CR
	DB	16,	3	; #
	DB	12,	20	; (
	DB	13,	30	; )
	DB	 0,	80	; *
	DB	 5,	70	; +
	DB	14,	10	; ,
	DB	 6,	70	; -
	DB	 1,	80	; /
	DB	16,	7	; A
	DB	16,	77H	; C
	DB	16,	5	; I
	DB	16,	55H	; T

GRP12	DB	16,	20H	; @A
	DB	22,	57H	; DA
	DB	17,	1	; DB
	DB	17,	2	; DS
	DB	17,	3	; DW
	DB	27,	0	; EN
	DB	16,	65H	; F0
	DB	16,	85H	; F1
	DB	17,	9	; IF
	DB	36,	0F0H	; IN
	DB	23,	0F6H	; JC
	DB	23,	0C6H	; JZ
	DB	10,	40	; OR
	DB	16,	19H	; P1
	DB	16,	1AH	; P2
	DB	16,	1CH	; P4
	DB	16,	1DH	; P5
	DB	16,	1EH	; P6
	DB	16,	1FH	; P7
	DB	16,	8	; R0
	DB	16,	9	; R1
	DB	16,	0AH	; R2
	DB	16,	0BH	; R3
	DB	16,	0CH	; R4
	DB	16,	0DH	; R5
	DB	16,	0EH	; R6
	DB	16,	0FH	; R7
	DB	22,	0E7H	; RL
	DB	22,	77H	; RR

GRP13	DB	16,	0	; @R0
	DB	16,	1	; @R1
	DB	20,	60H	; ADD
	DB	 9,	50	; AND
	DB	35,	50H	; ANL
	DB	16,	18H	; BUS
	DB	16,	75H	; CLK
	DB	25,	20H	; CLR
	DB	16,	45H	; CNT
	DB	25,	30H	; CPL
	DB	16,	32H	; DBB
	DB	34,	0	; DEC
	DB	28,	10H	; DIS
	DB	16,	0E5H	; DMA
	DB	17,	5	; END
	DB	17,	8	; EQU
	DB	19,	01H	; IDL
	DB	33,	10H	; INC
	DB	39,	0F0H	; INS
	DB	23,	12H	; JB0
	DB	23,	32H	; JB1
	DB	23,	52H	; JB2
	DB	23,	72H	; JB3
	DB	23,	92H	; JB4
	DB	23,	0B2H	; JB5
	DB	23,	0D2H	; JB6
	DB	23,	0F2H	; JB7
	DB	23,	0B6H	; JF0
	DB	23,	076H	; JF1
	DB	24,	04H	; JMP
	DB	23,	0E6H	; JNC
	DB	23,	86H	; JNI
	DB	23,	96H	; JNZ
	DB	23,	36H	; JT0
	DB	23,	56H	; JT1
	DB	23,	16H	; JTF
	DB	16,	0E5H	; MB0
	DB	16,	0F5H	; MB1
	DB	 2,	80	; MOD
	DB	44,	0	; MOV
	DB	19,	0	; NOP
	DB	 8,	60	; NOT
	DB	17,	11	; ORG
	DB	35,	40H	; ORL
	DB	37,	2	; OUT
	DB	16,	17H	; PSW
	DB	16,	0C5H	; RB0
	DB	16,	0D5H	; RB1
	DB	19,	83H	; RET
	DB	22,	0F7H	; RLC
	DB	22,	67H	; RRC
	DB	26,	0	; SEL
	DB	17,	13	; SET
	DB	 3,	80	; SHL
	DB	 4,	80	; SHR
	DB	16,	90H	; STS
	DB	20,	20H	; XCH
	DB	11,	40	; XOR
	DB	20,	0D0H	; XRL

GRP14	DB	20,	70H	; ADDC
	DB	32,	80H	; ANLD
	DB	24,	14H	; CALL
	DB	30,	0E0H	; DJNZ
	DB	17,	4	; ELSE
	DB	17,	7	; ENDM
	DB	29,	0	; ENT0
	DB	21,	93H	; JMPP
	DB	23,	26H	; JNT0
	DB	23,	46H	; JNT1
	DB	23,	86H	; JOBF
	DB	41,	20H	; MOVD
	DB	21,	83H	; MOVP
	DB	40,	80H	; MOVX
	DB	32,	70H	; ORLD
	DB	38,	20H	; OUTL
	DB	17,	12	; PAGE
	DB	19,	93H	; RETR
	DB	42,	30H	; STOP
	DB	43,	0	; STRT
	DB	22,	47H	; SWAP
	DB	16,	35H	; TCNT
	DB	31,	30H	; XCHD

GRP15	DB	17,	6	; ENDIF
	DB	16,	0F5H	; FLAGS
	DB	23,	0D6H	; JNIBF
	DB	17,	10	; MACRO
	DB	21,	0C3H	; MOVP3
	DB	16,	25H	; TCNTI
	DB	17,	14	; TITLE
;
; Evaluate expression to a 16 bit value, store in RESULT
;
EVAL:	XOR	AL,AL			; initialise
	MOV	BYTE PTR FNDEPTH,AL	; function stack
	MOV	BYTE PTR PRMDPTH,AL	; parameter stack
	DEC	AL
	MOV	BYTE PTR PFLAG,AL	; flag
	MOV	BX,0
	MOV	WORD PTR RESULT,BX	; result

; evaluation loop

EVAL1:	CALL	?EOX			; reached end of expression?
	JNZ	EVAL6
EVAL2:	MOV	BX,OFFSET FNDEPTH	; then perform any functions remaining
	MOV	AL,[BX]			; on stack
	OR	AL,AL
	JZ	EVAL3
	DEC	BYTE PTR [BX]
	MOV	DL,AL
	DEC	DL
	MOV	DH,0
	MOV	BX,OFFSET FUNCSTK
	ADD	BX,DX
	MOV	AL,[BX]
	CALL	DOFUNC
	JMP	EVAL2
					; eval completed, must have only a
EVAL3:	MOV	AL,BYTE PTR PRMDPTH	; single value left on parameter stack
	CMP	AL,2
	JZ	EVAL4
	CALL	EXPRERR
EVAL4:	CALL	?ERROR
	JZ	EVAL5
	RET

EVAL5:	MOV	BX,WORD PTR PRMSTK	; copy value and exit
	MOV	WORD PTR RESULT,BX
	RET

; begin evaluation

EVAL6:	CALL	?ERROR			; any errors so far?
	JZ	EVAL7
	JMP	EVAL36

EVAL7:	MOV	AL,BYTE PTR STRTYP	; get string type
	CMP	AL,QUOTE		; is it a quote?
	JNZ	EVAL11
	MOV	AL,BYTE PTR STRLEN	; then check length (one or two)
	OR	AL,AL
	JNZ	EVAL8
	CALL	EXPRERR
EVAL8:	CMP	AL,2+1
	JC	EVAL9
	CALL	EXPRERR
EVAL9:	MOV	DH,0
	MOV	BX,OFFSET STRING
	MOV	DL,[BX]			; get ascii value of first & second
	INC	BX			; characters
	DEC	AL
	JZ	EVAL10
	MOV	DH,[BX]
EVAL10:	XCHG	BX,DX
	JMP	EVAL34

EVAL11:	CMP	AL,BYTE PTR NUMER	; is it a number?
	JNZ	EVAL12
	MOV	BX,WORD PTR STRVAL	; then we already have its value
	JMP	EVAL34

EVAL12:	CALL	?MNEM			; is it a mnemonic?
	JZ	EVAL13
	JMP	EVAL29

EVAL13:	CMP	AL,16			; then check it is a function (0-15)
	JC	EVAL14
	JMP	EVAL27

EVAL14:	CMP	AL,12			; left parenthesis?
	MOV	CL,AL
	MOV	AL,BYTE PTR PFLAG
	JNZ	EVAL16
	OR	AL,AL
	JNZ	EVAL15
	CALL	EXPRERR
EVAL15:	MOV	AL,255
	MOV	BYTE PTR PFLAG,AL
	MOV	AL,CL
	JMP	EVAL21

EVAL16:	OR	AL,AL
	JNZ	EVAL23
EVAL17:	PUSH	CX			; precedence check
	MOV	AL,BYTE PTR FNDEPTH
	OR	AL,AL
	JZ	EVAL18
	MOV	DL,AL
	DEC	DL
	MOV	DH,0
	MOV	BX,OFFSET PRECSTK
	ADD	BX,DX
	MOV	AL,[BX]			; does function on stack
	CMP	AL,CH			; have a higher precedence?
	JC	EVAL18
	MOV	BX,OFFSET FNDEPTH
	MOV	[BX],DL
	MOV	BX,OFFSET FUNCSTK
	ADD	BX,DX
	MOV	AL,[BX]
	CALL	DOFUNC			; then perform it
	POP	CX
	JMP	EVAL17			; any more?

EVAL18:	POP	CX
	MOV	AL,CL
	CMP	AL,13			; right parenthesis
	JNZ	EVAL21
	MOV	BX,OFFSET FNDEPTH
	MOV	AL,[BX]
	OR	AL,AL
	JZ	EVAL19
	DEC	AL
	MOV	[BX],AL
	MOV	DL,AL
	MOV	DH,0
	MOV	BX,OFFSET FUNCSTK
	ADD	BX,DX
	MOV	AL,[BX]
	CMP	AL,12			; left parenthesis expected
	JZ	EVAL20
EVAL19:	CALL	EXPRERR
EVAL20:	XOR	AL,AL
	JMP	EVAL22

EVAL21:	CALL	PUSHFN			; push arithmetic function onto stack
	MOV	AL,255
EVAL22:	MOV	BYTE PTR PFLAG,AL
	JMP	EVAL36

EVAL23:	MOV	AL,CL
	CMP	AL,5			; +
	JZ	EVAL36
	CMP	AL,6			; if -
	JNZ	EVAL25
	INC	AL			; then make it unary
	MOV	CL,AL
	JMP	EVAL17

EVAL25:	CMP	AL,8			; not
	JZ	EVAL26
	CALL	EXPRERR
EVAL26:	JMP	EVAL17

; not a function, so just use the value from the mnemonic tables

EVAL27:	CMP	AL,17			; but psuedo-ops not allowed
	JNZ	EVAL28
	CALL	EXPRERR
EVAL28:	MOV	BL,CH
	MOV	BH,0
	JMP	EVAL34

EVAL29:	MOV	AL,BYTE PTR STRTYP 	; is it '$'
	CMP	AL,OTHER
	JNZ	EVAL31
	MOV	AL,BYTE PTR STRING
	CMP	AL,'$'
	JZ	EVAL30
	CALL	EXPRERR
	MOV	BX,0
	JMP	EVAL34

EVAL30:	MOV	BX,WORD PTR PC		; then get current program counter
	JMP	EVAL34

; must be symbol

EVAL31:	CALL	FNDSYM
	CALL	?SYM
	JNZ	EVAL32
	CALL	PHASERR			; not previously encountered
	CALL	ADDSYM
	JMP	EVAL33

EVAL32:	CALL	GETNBL			; value defined?
	AND	AL,00000111B
	JNZ	EVAL33
	CALL	UDEFERR			; no
EVAL33:	CALL	GETSVAL			; get value from symbol table

EVAL34:	MOV	AL,BYTE PTR PFLAG	; push (BX) onto param stack
	OR	AL,AL
	JNZ	EVAL35
	CALL	EXPRERR
EVAL35:	XOR	AL,AL
	MOV	BYTE PTR PFLAG,AL
	CALL	PSHPRM

EVAL36:	CALL	PARSE			; get next string
	JMP	EVAL1			; and loop again
;
; Test for end of expression
;
?EOX:	MOV	AL,BYTE PTR STRTYP
	CMP	AL,OTHER
	JZ	?EOX1
	RET

?EOX1:	MOV	AL,BYTE PTR STRING
	CMP	AL,CR
	JNZ	?EOX2
	RET

?EOX2:	CMP	AL,';'
	JNZ	?EOX3
	RET

?EOX3:	CMP	AL,','
	JNZ	?EOX4
	RET

?EOX4:	CMP	AL,'!'
	RET
;
; Push parameter BX onto parameter stack
;
PSHPRM:	XCHG	BX,DX
	MOV	BX,OFFSET PRMDPTH
	MOV	AL,[BX]
	CMP	AL,16
	JC	PSHPRM1

	CALL	EXPRERR			; overflow
	MOV	BYTE PTR [BX],0
PSHPRM1:MOV	AL,[BX]
	INC	BYTE PTR [BX]
	INC	BYTE PTR [BX]
	MOV	CL,AL
	MOV	CH,0
	MOV	BX,OFFSET PRMSTK
	ADD	BX,CX
	MOV	[BX],DL
	INC	BX
	MOV	[BX],DH
	RET
;
; Push arithmetic function code A, and precedence B, onto stacks
;
PUSHFN:	LAHF
	PUSH	AX
	MOV	BX,OFFSET FNDEPTH
	MOV	AL,[BX]
	CMP	AL,10
	JC	PUSHFN1

	MOV	BYTE PTR [BX],0		; overflow
	CALL	EXPRERR
PUSHFN1:MOV	DL,[BX]
	MOV	DH,0
	INC	BYTE PTR [BX]
	POP	AX
	SAHF
	MOV	BX,OFFSET FUNCSTK	; function stack
	ADD	BX,DX
	MOV	[BX],AL
	MOV	BX,OFFSET PRECSTK	; precedence stack
	ADD	BX,DX
	MOV	[BX],CH
	RET
;
; Pop parameter BX from parameter stack
;
POPPRM:	MOV	BX,OFFSET PRMDPTH
	MOV	AL,[BX]
	OR	AL,AL
	JNZ	POPPRM1

	CALL	EXPRERR			; underflow
	MOV	BX,0
	RET

POPPRM1:DEC	BYTE PTR [BX]
	DEC	BYTE PTR [BX]
	MOV	CL,[BX]
	MOV	CH,0
	MOV	BX,OFFSET PRMSTK
	ADD	BX,CX
	MOV	CL,[BX]
	INC	BX
	MOV	BH,[BX]
	MOV	BL,CL
	RET
;
; Get two 16 bit values from evaluation stack
;
GET2:	CALL	POPPRM
	XCHG	BX,DX
	JMP	POPPRM
;
; Same as GET2 but check DE is less than 16
;
GET2A:	CALL	GET2
	MOV	AL,DH
	OR	AL,AL
	JNZ	GET2B

	MOV	AL,DL
	CMP	AL,16+1
	JNC	GET2B
	RET

GET2B:	CALL	EXPRERR
	MOV	AL,16
	RET
;
; Perform arithmetic function
;
DOFUNC:	MOV	BL,AL
	MOV	BH,0
	ADD	BX,BX
	MOV	DX,OFFSET FUNCTBL
	ADD	BX,DX
	MOV	DL,[BX]
	INC	BX
	MOV	BH,[BX]
	MOV	BL,DL
	JMP	BX

FUNCTBL	DW	STAR		; *
	DW	SLASH		; /
	DW	MODD		; MOD
	DW	SHLL		; SHL
	DW	SHRR		; SHR
	DW	PLUS		; +
	DW	MINUS		; -
	DW	NEGG		; unary minus
	DW	NOTT		; NOT
	DW	ANDD		; AND
	DW	ORR		; OR
	DW	XORR		; XOR

	DW	EXPRERR		;
;
; Perform negation	BX <- 0-BX
;
NEGATE:	XOR	AL,AL
	SUB	AL,BL
	MOV	BL,AL
	MOV	AL,0
	SBB	AL,BH
	MOV	BH,AL
	RET
;
; Pertform division
;
DIV1:	CALL	GET2

; divide BX by DE, quot=DE  mod=BX

DIV2:	XCHG	BX,DX
	MOV	WORD PTR DIVTMP,BX	; temp save
	MOV	BX,OFFSET DIVCNT
	MOV	BYTE PTR [BX],16+1 	; number of loops
	MOV	CX,OFFSET 0
	PUSH	CX
	XOR	AL,AL
DIV3:	MOV	AL,DL
	RCL	AL,1
	MOV	DL,AL
	MOV	AL,DH
	RCL	AL,1
	MOV	DH,AL
	DEC	BYTE PTR [BX]
	POP	BX
	JNZ	DIV31
	RET

DIV31:	MOV	AL,0
	ADC	AL,0
	ADD	BX,BX
	MOV	CH,BH
	ADD	AL,BL
	MOV	BX,WORD PTR DIVTMP
	SUB	AL,BL
	MOV	CL,AL
	MOV	AL,CH
	SBB	AL,BH
	MOV	CH,AL
	PUSH	CX
	JNC	DIV4

	ADD	BX,CX
	XCHG	SP,BP
	XCHG	[BP],BX
	XCHG	SP,BP
DIV4:	MOV	BX,OFFSET DIVCNT
	CMC
	JMP	DIV3
;
; Perform multiplication  BX <-  BX * DE
;
MULT:	MOV	CH,BH
	MOV	CL,BL
	MOV	BX,0
MULT1:	XOR	AL,AL
	MOV	AL,CH
	RCR	AL,1
	MOV	CH,AL
	MOV	AL,CL
	RCR	AL,1
	MOV	CL,AL
	JC	MULT2
	OR	AL,CH
	JNZ	MULT3
	RET

	JMP	MULT3

MULT2:	ADD	BX,DX
MULT3:	XCHG	BX,DX
	ADD	BX,BX
	XCHG	BX,DX
	JMP	MULT1
;
; Multiply   *
;
STAR:	CALL	GET2
	CALL	MULT
	JMP	PSHPRM
;
; Divide   /
;
SLASH:	CALL	DIV1
	XCHG	BX,DX
	JMP	PSHPRM
;
; Modulus   MOD
;
MODD:	CALL	DIV1
	JMP	PSHPRM
;
; Shift left   SHL
;
SHLL:	CALL	GET2A
SHLL1:	OR	AL,AL
	JNZ	SHLL2
	JMP	PSHPRM

SHLL2:	ADD	BX,BX
	DEC	AL
	JMP	SHLL1
;
; Shift right	SHR
;
SHRR:	CALL	GET2A
SHRR1:	OR	AL,AL
	JNZ	SHRR2
	JMP	PSHPRM

SHRR2:	LAHF
	PUSH	AX
	XOR	AL,AL
	MOV	AL,BH
	RCR	AL,1
	MOV	BH,AL
	MOV	AL,BL
	RCR	AL,1
	MOV	BL,AL
	POP	AX
	SAHF
	DEC	AL
	JMP	SHRR1
;
; Addition   +
;
PLUS:	CALL	GET2
PLUS1:	ADD	BX,DX
	JMP	PSHPRM
;
; Subtraction	-
;
MINUS:	CALL	GET2
	XCHG	BX,DX
	CALL	NEGATE
	JMP	PLUS1
;
; Unary minus
;
NEGG:	CALL	POPPRM
NEGG1:	CALL	NEGATE
	JMP	PSHPRM
;
; Two's complement
;
NOTT:	CALL	POPPRM
	INC	BX
	JMP	NEGG1
;
; Logical AND
;
ANDD:	CALL	GET2
	MOV	AL,DH
	AND	AL,BH
	MOV	BH,AL
	MOV	AL,DL
	AND	AL,BL
	MOV	BL,AL
	JMP	PSHPRM
;
; Logical OR
;
ORR:	CALL	GET2
	MOV	AL,DH
	OR	AL,BH
	MOV	BH,AL
	MOV	AL,DL
	OR	AL,BL
	MOV	BL,AL
	JMP	PSHPRM
;
; Logical XOR
;
XORR:	CALL	GET2
	MOV	AL,DH
	XOR	AL,BH
	MOV	BH,AL
	MOV	AL,DL
	XOR	AL,BL
	MOV	BL,AL
	JMP	PSHPRM
;
;	Begin assembly
;
PASS1:	XOR	AL,AL
	MOV	BYTE PTR PASS,AL	; first pass
	CALL	CLRSYM			; clear all symbols

PASS2:	CALL	INITLN			; initialise for next line
	CALL	OPENSRC			; open source file

	MOV	BX,0			; initialise -
	MOV	WORD PTR LABLPTR,BX	; labels
	MOV	WORD PTR PC,BX		; program counter
	MOV	WORD PTR CURPC,BX	; current program counter
	MOV	WORD PTR EXEC,BX	; execution address
	XOR	AL,AL
	MOV	BYTE PTR NESTLVL,AL	; nested IF's
	MOV	BYTE PTR HEXREC,AL	; hex record is data

	MOV	AL,BYTE PTR PAGLEN	; output first header
	MOV	BYTE PTR LINES,AL
	CALL	HEADER

; process next assembler statement

NXTSTM:	CALL	PARSE			; parse next string
NXTSTM1:MOV	AL,BYTE PTR STRTYP	; check type

	CMP	AL,NUMER		; ignore line numbers
	JZ	NXTSTM

	CMP	AL,QUOTE		; quotes not allowed here
	JNZ	NXTSTM2
	JMP	SYNTERR

NXTSTM2:CMP	AL,ALPHA		; label or mnemonic?
	JZ	NXTSTM5

	MOV	AL,BYTE PTR STRING	; if full line comment,
	CMP	AL,'*'
	JZ	NXTSTM3
	JMP	ENDSTM

NXTSTM3:CALL	?LABEL			; no labels must be present
	JZ	NXTSTM4
	JMP	SYNTERR

NXTSTM4:JMP	SKIP			; skip remainder of line

; label or mnemonic

NXTSTM5:CALL	?MNEM			; if a mnemonic
	JNZ	NXTSTM6
	JMP	DOMNEM			; process it including any operands

NXTSTM6:CALL	FNDSYM			; must be a label, search table
	CALL	?SYM
	JNZ	NXTSTM8

	CALL	ADDSYM			; label not found, add symbol to table
	MOV	AL,BYTE PTR PASS 	; this shouldn't happen on pass 2
	OR	AL,AL
	JZ	NXTSTM7
	CALL	PHASERR
NXTSTM7:JMP	NXTSTM9

NXTSTM8:CALL	GETNBL			; label found
	CMP	AL,00000110B
	JNZ	NXTSTM9

	CALL	NIMERR
	JMP	SKIP

NXTSTM9:MOV	BX,WORD PTR LABLPTR	; check for multiple labels
	MOV	AL,BL
	OR	AL,BH
	JZ	NXTSTM10
	CALL	LABLERR
NXTSTM10:
	MOV	BX,WORD PTR SYMPTR	; assign this symbol as a label
	MOV	WORD PTR LABLPTR,BX

	CALL	PARSE			; labels may include a trailing ':'
	MOV	AL,BYTE PTR STRTYP	; so check if one is present
	CMP	AL,OTHER
	JNZ	NXTSTM1

	MOV	AL,BYTE PTR STRING
	CMP	AL,':'
	JNZ	NXTSTM1

	JMP	NXTSTM			; loop
;
; Reached end of statement
;
ENDSTM:	CALL	LABELL
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,OTHER		; no more numbers, quotes or alpha
	JZ	ENDSTM1
	JMP	SYNTERR			; strings are allowed

ENDSTM1:MOV	AL,BYTE PTR STRING
	CMP	AL,CR			; CR?
	JNZ	ENDSTM2

	CALL	PARSE			; then consume LF
	JMP	NXTSTM

ENDSTM2:CMP	AL,';'			; comment?
	JNZ	ENDSTM3

	CALL	LABELL
	JMP	SKIP

ENDSTM3:CMP	AL,'!'			; statement separator?
	JNZ	ENDSTM4
	JMP	NXTSTM

ENDSTM4:CMP	AL,EOF			; end of file?
	JZ	NXTPAS
	CMP	AL,NULL
	JZ	NXTPAS

	JMP	SYNTERR
;
; Skip remainder until end of line or next statement
;
SKIP:	CALL	PARSE			; parse next string
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,OTHER
	JNZ	SKIP			; loop until terminator found

	MOV	AL,BYTE PTR STRING
	CMP	AL,'!'			; statement separator?
	JNZ	SKIP1
	JMP	NXTSTM

SKIP1:	CMP	AL,LF			; line feed?
	JNZ	SKIP2
	JMP	NXTSTM

SKIP2:	CMP	AL,EOF			; end-of-file?
	JZ	NXTPAS
	CMP	AL,NULL
	JZ	NXTPAS

	JMP	SKIP
;
; Next pass
;
NXTPAS:	MOV	AL,BYTE PTR PASS	; increment pass counter
	INC	AL
	MOV	BYTE PTR PASS,AL
	CMP	AL,1			; was it only first pass?
	JNZ	NXTPAS1

	CALL	SELSRC			; close source file
	MOV	DX,OFFSET SRCFCB
	CALL	CLOSEF
	JMP	PASS2			; then perform second pass

NXTPAS1:CALL	PARSE

	MOV	BX,WORD PTR ERRCNT	; any errors during assembly?
	MOV	AL,BH
	OR	AL,BL
	LAHF
	PUSH	AX			; save flag
	JNZ	NXTPAS2

	MOV	BYTE PTR PRC,0		; no, clear return code

NXTPAS2:CALL	WRTCR			; send err count msg to file
	MOV	BX,OFFSET MSG11
	CALL	WRTSTR
	POP	DX			; restore flag

	MOV	AL,BYTE PTR PRN		; also to screen if no screen list
	CMP	AL,'X'-'A'
	JNZ	NXTPAS3
	JMP	DONE

NXTPAS3:MOV	AL,DH			; if errors, turn up one line
	OR	AL,AL
	JZ	NXTPAS4
	CALL	CRLF

NXTPAS4:MOV	BX,OFFSET MSG11
	CALL	PRINT
	JMP	DONE			; close all files and exit
;
; Handle mnemonics/assembler directives
;
DOMNEM:	CMP	AL,17			; psuedo op?
	JZ	DOMNEM1
	JMP	DOOPC

DOMNEM1:MOV	DL,CH
	MOV	DH,0
	DEC	DX
	MOV	BX,OFFSET PSTBL
	ADD	BX,DX
	ADD	BX,DX
	MOV	DL,[BX]
	INC	BX
	MOV	BH,[BX]
	MOV	BL,DL
	JMP	BX

; Table of psuedo-opcodes

PSTBL	DW	DBB		; DB
	DW	DSS		; DS
	DW	DWW		; DW
	DW	ELSEE		; ELSE
	DW	ENDD		; END
	DW	ENDIFF		; ENDIF
	DW	NIMM		; ENDM
	DW	EQUU		; EQU
	DW	IFF		; IF
	DW	NIMM		; MACRO
	DW	ORGG		; ORG
	DW	PAGEE		; PAGE
	DW	SETT		; SET
	DW	TITLEE		; TITLE
;
; Define byte  DB
;
DBB:	CALL	LABELL
DBB1:	CALL	PARSE
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,QUOTE
	JNZ	DBB4

	MOV	AL,BYTE PTR STRLEN ; quote
	DEC	AL
	JZ	DBB4

	MOV	CH,AL
	INC	CH
	INC	CH
	MOV	BX,OFFSET STRING
DBB2:	DEC	CH
	JZ	DBB3

	PUSH	CX
	MOV	AL,[BX]
	INC	BX
	PUSH	BX
	CALL	OUTBYT
	POP	BX
	POP	CX
	JMP	DBB2

DBB3:	CALL	PARSE
	JMP	DBB5

DBB4:	CALL	EVALW
	MOV	AL,BH
	OR	AL,AL
	JZ	DBB41
	CALL	DATAERR

DBB41:	MOV	AL,BL
	CALL	OUTBYT
DBB5:	CALL	ALIGNPC
	CALL	?SEP
	CMP	AL,','
	JZ	DBB1

	JMP	ENDSTM
;
; Define storage  DS
;
DSS:	CALL	LABELL
	CALL	INSPC
	CALL	GETOPR
	XCHG	BX,DX
	MOV	BX,WORD PTR PC
	ADD	BX,DX
	MOV	WORD PTR PC,BX
	MOV	WORD PTR CURPC,BX
	JMP	ENDSTM
;
; Define word  DW
;
DWW:	CALL	LABELL
DWW1:	CALL	GETOPR
	PUSH	BX
	MOV	AL,BH			; output high byte (MSB) first
	CALL	OUTBYT
	POP	BX
	MOV	AL,BL			; then output low byte (LSB)
	CALL	OUTBYT
	CALL	ALIGNPC
	CALL	?SEP
	CMP	AL,','
	JZ	DWW1

	JMP	ENDSTM
;
; Conditional  ELSE
;
ELSEE:	CALL	LABELL
	CALL	UNNEST
	CMP	AL,1
	MOV	AL,2
	JNZ	ELSEE1
	JMP	IFF3

ELSEE1:	CALL	BALERR
	JMP	SKIPOPR
;
; End assembly  END
;
ENDD:	CALL	LABELL
	CALL	INSPC
	CALL	?ERROR			; error flagged?
	JZ	ENDD1
	JMP	ENDSTM

ENDD1:	CALL	PARSE			; test for operand
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,OTHER
	JZ	ENDD2

	CALL	EVALW			; evaluate
	MOV	WORD PTR EXEC,BX
ENDD2:	MOV	AL,BYTE PTR NESTLVL	; conditionals should balance
	OR	AL,AL
	JZ	ENDD3
	CALL	BALERR
ENDD3:	CALL	PARSE
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,OTHER
	JZ	ENDD4
	JMP	SYNTERR

ENDD4:	MOV	AL,BYTE PTR STRING
	CMP	AL,LF
	JZ	ENDD5
	JMP	SYNTERR

ENDD5:	JMP	NXTPAS

;
; End conditional  ENDIF
;
ENDIFF:	CALL	LABELL
	CALL	UNNEST
	JMP	SKIPOPR
;
; Equate  EQU
;
EQUU:	CALL	?LABEL
	JNZ	EQUU1
	JMP	SYNTERR

EQUU1:	MOV	BX,WORD PTR PC
	PUSH	BX
	CALL	GETOPR
	MOV	WORD PTR PC,BX
	CALL	LABELL
	CALL	INSWRD
	MOV	BX,OFFSET PCFLD+5	; symbol value
	MOV	BYTE PTR [BX],'='
	POP	BX
	MOV	WORD PTR PC,BX
	JMP	ENDSTM
;
; Begin conditional  IF
;
IFF:	CALL	LABELL
	CALL	GETOPR			; get operand
	CALL	?ERROR			; if error flagged
	JNZ	IFF3			; then regard conditional as false

	MOV	AL,BYTE PTR ASMTST	; using ASM style IF test?
	OR	AL,AL
	MOV	AL,BL
	JZ	IFF1
	AND	AL,00000001B		; DRI assemblers test bit 0 only
	JMP	IFF2

IFF1:	OR	AL,BH			; others test all bits
IFF2:	MOV	AL,1
	JZ	IFF3

	CALL	NEST			; true condition
	JMP	ENDSTM

; Skip while IF/ELSE false

IFF3:	MOV	BYTE PTR IFMODE,AL
	XOR	AL,AL
	MOV	BYTE PTR IFCOUNT,AL

IFF4:	MOV	AL,BYTE PTR STRTYP	; parse statement
	CMP	AL,OTHER
	JNZ	IFF6

	MOV	AL,BYTE PTR STRING
	CMP	AL,CR
	JNZ	IFF5
	CALL	PARSE
	JMP	IFF7

IFF5:	CMP	AL,'!'
	JZ	IFF7

	CMP	AL,EOF			; premature end of file?
	JZ	IFF61
	CMP	AL,NULL
	JZ	IFF61

IFF6:	CALL	PARSE
	JMP	IFF4

IFF61:	CALL	BALERR
	JMP	NXTPAS

IFF7:	CALL	PARSE			; reached end of statement
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,NUMER		; ignore line number
	JNZ	IFF8
	CALL	PARSE
IFF8:	MOV	AL,BYTE PTR STRTYP
	CMP	AL,ALPHA
	JNZ	IFF4			; loop if not alpha

	CALL	?MNEM			; if mnemonic,
	JZ	IFF10			; check it

	CALL	PARSE
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,OTHER
	JNZ	IFF9
	MOV	AL,BYTE PTR STRING
	CMP	AL,':'
	JNZ	IFF4
	CALL	PARSE
IFF9:	MOV	AL,BYTE PTR STRTYP
	CMP	AL,ALPHA
	JNZ	IFF4
	CALL	?MNEM
	JNZ	IFF4

IFF10:	CMP	AL,17			; found mnemonic, what type?
	JNZ	IFF4
	MOV	AL,CH
	CMP	AL,9			; is it 'IF'
	JNZ	IFF12
	MOV	BX,OFFSET IFCOUNT	; then increment count
	INC	BYTE PTR [BX]		; check for overflow
	JNZ	IFF11
	CALL	OVFLERR
IFF11:	JMP	IFF4			; and keep looping

IFF12:	CMP	AL,4			; is it 'ELSE'
	JNZ	IFF15
	MOV	AL,BYTE PTR IFCOUNT	; ignore it?
	OR	AL,AL
	JZ	IFF13
	JMP	IFF4

IFF13:	MOV	AL,BYTE PTR IFMODE	; no, then start assembling
	CMP	AL,2
	JNZ	IFF14
	CALL	BALERR
IFF14:	MOV	AL,2
	CALL	NEST
	JMP	SKIPOPR

IFF15:	CMP	AL,6			; is it 'ENDIF'
	JZ	IFF16
	JMP	IFF4

IFF16:	MOV	BX,OFFSET IFCOUNT	; then decrement 'IF' count
	MOV	AL,[BX]
	DEC	BYTE PTR [BX]
	OR	AL,AL
	JZ	IFF17
	JMP	IFF4

IFF17:	JMP	SKIPOPR			; if reached zero, start assembling

NEST:	MOV	CH,AL
	MOV	BX,OFFSET NESTLVL
	MOV	AL,[BX]
	CMP	AL,8			; max nest level
	JC	NEST1
	JMP	OVFLERR

NEST1:	INC	BYTE PTR [BX]
	MOV	DL,AL
	MOV	DH,0
	MOV	BX,OFFSET NESTSTK
	ADD	BX,DX
	MOV	[BX],CH
	RET

UNNEST:	MOV	BX,OFFSET NESTLVL
	MOV	AL,[BX]
	OR	AL,AL
	JNZ	UNNEST1
	JMP	BALERR

UNNEST1:DEC	BYTE PTR [BX]
	MOV	DL,[BX]
	MOV	DH,0
	MOV	BX,OFFSET NESTSTK
	ADD	BX,DX
	MOV	AL,[BX]
	RET
;
; Assembly origin  ORG
;
ORGG:	CALL	GETOPR			; get operand, skip if error
	CALL	?ERROR
	JZ	ORGG1
	JMP	ENDSTM

ORGG1:	MOV	WORD PTR PC,BX
	MOV	WORD PTR CURPC,BX
	CALL	LABELL
	CALL	INSPC
	JMP	ENDSTM
;
; Page  PAGE
;
PAGEE:	CALL	PARSE			; check for operand
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,OTHER
	JZ	PAGEE1

	CALL	EVALB			; set page length
	MOV	BYTE PTR PAGLEN,AL
	JMP	ENDSTM

PAGEE1:	MOV	AL,BYTE PTR PAGLEN	; force page on next line
	DEC	AL
	MOV	BYTE PTR LINES,AL
	JMP	ENDSTM
;
; Assign value  SET
;
SETT:	CALL	?LABEL
	JNZ	SETT1
	JMP	SYNTERR

SETT1:	CALL	GETNBL
	OR	AL,AL			; test if first encounter
	JZ	SETT2

	CMP	AL,00000101B
	JZ	SETT2
	CALL	LABLERR
SETT2:	MOV	AL,00000101B
	CALL	PUTNBL
	CALL	GETOPR
	PUSH	BX
	CALL	?LABEL
	POP	BX
	CALL	PUTSVAL
	MOV	BX,0			; inhibit 'phase error' (pc <> label)
	MOV	WORD PTR LABLPTR,BX
	JMP	ENDSTM
;
; Title  TITLE
;
TITLEE:	CALL	PARSE
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,QUOTE
	JZ	TITLEE1
	JMP	SYNTERR

TITLEE1:MOV	BX,OFFSET STRLEN	; move string to title field
	MOV	CH,[BX]
	INC	BX
	MOV	DX,OFFSET MSG12B
	CALL	MOVE
	XCHG	BX,DX			; place trailing null
	MOV	BYTE PTR [BX],0
	CALL	PARSE
	JMP	ENDSTM
;
; Functions not implemented
;
NIMM:	CALL	NIMERR
	JMP	SKIPOPR
;
; Skip operand and rest of statement
;
SKIPOPR:CALL	PARSE
	JMP	ENDSTM
;
;	8048/41 opcodes
;
DOOPC:	SUB	AL,19			; modes 19-44
	CMP	AL,(44-19)+1
	JC	DOOPC1
	JMP	SYNTERR

DOOPC1:	MOV	DL,AL
	MOV	DH,0
	MOV	BX,OFFSET MDTBL
	ADD	BX,DX
	ADD	BX,DX
	MOV	DL,[BX]
	INC	BX
	MOV	BH,[BX]
	MOV	BL,DL
	JMP	BX
;
; Mode table
; B contains a value which is (or is used to determine) the opcode
;
MDTBL	DW	MD19		; implied
	DW	MD20		; ADD ADDC XRL XCH
	DW	MD21		; JMPP MOVP MOVP3
	DW	MD22		; accumulator
	DW	MD23		; short jump
	DW	MD24		; JMP CALL
	DW	MD25		; CLR CPL
	DW	MD26		; SEL
	DW	MD27		; EN
	DW	MD28		; DIS
	DW	MD29		; ENT0
	DW	MD30		; DJNZ
	DW	MD31		; XCHD
	DW	MD32		; ANLD ORLD
	DW	MD33		; INC
	DW	MD34		; DEC
	DW	MD35		; ORL ANL
	DW	MD36		; IN
	DW	MD37		; OUT
	DW	MD38		; OUTL
	DW	MD39		; INS
	DW	MD40		; MOVX
	DW	MD41		; MOVD
	DW	MD42		; STOP
	DW	MD43		; STRT
	DW	MD44		; MOV
;
; Implied - no operand
;
MD19:	CALL	OUTBYT1
	CALL	PARSE
	JMP	ENDOPR
;
; ADD, ADDC, XRL, XCH
;
MD20:	CALL	ACOMMA			; A,
	CALL	OPRREG
;
; #, @Ri, Rr
;
MD201:	CMP	AL,3			; #
	JZ	MD202
	JMP	DORR

MD202:	MOV	AL,CH
	CMP	AL,20H			; XCH
	JNZ	MD203
	CALL	REGERR

MD203:	AND	AL,11101111B		; ADD, ADDC
	CMP	AL,01100000B
	JNZ	MD204

	MOV	AL,CH
	SUB	AL,60H
	MOV	CH,AL
MD204:	MOV	AL,3
MD205:	ADD	AL,CH
	CALL	OUTBYT			; opcode
	CALL	EVALB
	CALL	OUTBYT			; operand
	JMP	ENDOPR
;
; @Ri, Rr
;
DORR:	CALL	?RR			; Rr
	JZ	DORR3

DORR1:	CALL	?RI			; @Ri
DORR2:	JZ	DORR3
	CALL	REGERR
DORR3:	ADD	AL,CH
	CALL	OUTBYT
	JMP	ENDOPR
;
; JMPP, MOVP, MOVP3
;
MD21:	MOV	AL,CH
	CMP	AL,93H			; JMPP
	JZ	MD211

	CALL	ACOMMA
MD211:	CALL	OPRREG
	CMP	AL,20H			; @A
	JMP	DORR2
;
; Accumulator
;
MD22:	CALL	OUTBYT1
MD221:	CALL	OPRREG
	CMP	AL,7			; A
	JZ	MD222
	CALL	REGERR
MD222:	JMP	ENDOPR

;
; Short jump
;
MD23:	CALL	OUTBYT1
MD231:	CALL	GETOPR			; addr
	MOV	AL,BYTE PTR PC+1
	CMP	AL,BH			; same page?
	PUSH	BX
	JZ	MD232
	CALL	VALUERR

MD232:	POP	BX
	MOV	AL,BL
	CALL	OUTBYT
	JMP	ENDOPR
;
; CALL, JMP
;
MD24:	CALL	GETOPR
	MOV	AL,BYTE PTR PC+1
	XOR	AL,BH
	AND	AL,11111000B		; address within 2k boundary?
	JZ	MD241
	CALL	VALUERR

MD241:	MOV	AL,BH
	ADD	AL,AL
	ROL	AL,1
	ROL	AL,1
	ROL	AL,1
	ROL	AL,1
	OR	AL,CH
	PUSH	BX
	CALL	OUTBYT			; opcode
	POP	BX
	MOV	AL,BL
	CALL	OUTBYT			; operand
	JMP	ENDOPR
;
; CLR, CPL
;
MD25:	CALL	OPRREG
	MOV	BX,OFFSET REGS1
	MOV	CL,4
MD251:	CALL	SCAN
	JMP	DORR2
;
; SEL
;
MD26:	CALL	OPRREG
	MOV	BX,OFFSET REGS2
	MOV	CL,4
	JMP	MD251
;
; EN
;
MD27:	CALL	OPRREG
	MOV	BX,OFFSET REGS3
	MOV	CL,4
	JMP	MD251
;
; DIS
;
MD28:	CALL	OPRREG
	MOV	BX,OFFSET REGS4
	MOV	CL,2
	JMP	MD251

REGS1	DB	07H,77H,65H,85H		; A C F0 F1
REGS2	DB	0C5H,0D5H		; RB0 RB1
REGS3	DB	0E5H,0F5H		; MB0/DMA MB1/FLAGS
REGS4	DB	05H,25H			; I TCNTI

;
; ENT0
;
MD29:	CALL	OPRREG
	CMP	AL,75H
	JMP	DORR2
;
; DJNZ
;
MD30:	CALL	OPRREG
	CALL	?RR			; test R0-R7
	JZ	MD301
	CALL	REGERR

MD301:	OR	AL,CH
	CALL	OUTBYT
	CALL	?COMMA			; comma
	JMP	MD231			; addr
;
; XCHD
;
MD31:	CALL	ACOMMA			; A,@Ri
	CALL	OPRREG
	JMP	DORR1
;
; ANLD, ORLD
;
MD32:	CALL	OPRREG
MD321:	CALL	?PP4			; Pp4,A
MD322:	JZ	MD323
	CALL	REGERR
MD323:	ADD	AL,CH
	CALL	OUTBYT
	CALL	?COMMA
	JMP	MD221
;
; INC
;
MD33:	CALL	OPRREG
	CMP	AL,7
	JNZ	MD331
	JMP	DORR3

MD331:	CALL	?RRI
	JNZ	MD332
	JMP	DORR3

MD332:	CALL	?RR
	JMP	DORR2
;
; DEC
;
MD34:	CALL	OPRREG
	CMP	AL,7
	JNZ	MD341
	JMP	DORR3

MD341:	MOV	CH,0C0H
	JMP	MD332
;
; ANL ORL
;
MD35:	CALL	OPRREG
	CMP	AL,18H			; BUS
	JZ	MD352

	CALL	?PP1			; P1 P2
	JZ	MD352

	CMP	AL,7			; A
	JZ	MD351
	CALL	REGERR

MD351:	CALL	?COMMA
	CALL	OPRREG
	JMP	MD201

MD352:	ADD	AL,CH
	ADD	AL,30H
	CALL	OUTBYT
	CALL	?COMMA
	CALL	OPRREG
	CMP	AL,3			; #
	JZ	MD353
	CALL	REGERR

MD353:	CALL	EVALB
	CALL	OUTBYT
	JMP	ENDOPR
;
; IN
;
MD36:	CALL	ACOMMA
	CALL	OPRREG
	CALL	?PP1
	JNZ	MD361
	JMP	DORR3

MD361:	CMP	AL,32H
	JMP	DORR2
;
; OUT
;
MD37:	CALL	OPRREG
	CMP	AL,32H
	MOV	AL,0
	JMP	MD322
;
; OUTL
;
MD38:	CALL	OPRREG
	CALL	?PP1
	JNZ	MD381
	JMP	MD323

MD381:	CMP	AL,18H
	MOV	CH,0EAH
	JMP	MD322
;
; INS
;
MD39:	CALL	ACOMMA
	CALL	OPRREG
	CMP	AL,18H
	JMP	DORR2
;
; MOVX
;
MD40:	CALL	OPRREG
	CMP	AL,7
	JZ	MD401

	MOV	CH,90H
	CALL	?RRI			; @Ri,A
	JMP	MD322

MD401:	CALL	?COMMA
	CALL	OPRREG
	JMP	DORR1
;
; MOVD
;
MD41:	CALL	OPRREG
	CMP	AL,7
	JZ	MD411
	JMP	MD321

MD411:	MOV	CH,0F0H
	CALL	?COMMA			; ,Pp4
	CALL	OPRREG
	CALL	?PP4
	JMP	DORR2
;
; STOP
;
MD42:	CALL	OPRREG
	CMP	AL,35H			; TCNT
	JMP	DORR2
;
; STRT
;
MD43:	CALL	OPRREG
	CMP	AL,45H			; CNT
	JNZ	MD431
	JMP	DORR3

MD431:	CMP	AL,55H			; T
	JMP	DORR2
;
; MOV
;
MD44:	CALL	OPRREG
	CMP	AL,7			; A,
	JNZ	MD444

	CALL	?COMMA
	CALL	OPRREG
	CALL	?RRI			; @Ri Rr
	JNZ	MD441

	MOV	CH,0F0H
	JMP	DORR3

MD441:	CMP	AL,3			; #
	JNZ	MD442

	MOV	CH,20H
	JMP	MD204

MD442:	CMP	AL,17H			; PSW
	JNZ	MD443

	MOV	CH,0B0H
	JMP	DORR3

MD443:	CMP	AL,55H			; T
	MOV	CH,0EDH
	JMP	DORR2

MD444:	CALL	?RRI
	JNZ	MD446
;
	MOV	CH,AL
	CALL	?COMMA			; @Ri, Rr,
	CALL	OPRREG
	CMP	AL,3			; @Ri,#  Rr,#
	JNZ	MD445

	MOV	AL,0B0H
	JMP	MD205

MD445:	CMP	AL,7			; @Ri,A  Rr,A
	MOV	AL,0A0H
	JMP	DORR2
;
MD446:	CMP	AL,17H			; PSW,A
	JNZ	MD447

	MOV	CH,0C0H
	JMP	MD323

MD447:	CMP	AL,90H			; STS,A
	JNZ	MD448
	JMP	MD323

MD448:	CMP	AL,55H			; T,A
	MOV	CH,0DH
	JMP	MD322
;
; Test for a valid data separator or end of statement
;
?SEP:	MOV	AL,BYTE PTR STRTYP
	CMP	AL,OTHER
	JZ	?SEP1
	CALL	DATAERR

?SEP1:	MOV	AL,BYTE PTR STRING
	CMP	AL,','			; data separator
	JNZ	?SEP2
	RET

?SEP2:	CMP	AL,';'			; begining of comment
	JNZ	?SEP3
	RET

?SEP3:	CMP	AL,CR
	JZ	?SEP4
	CALL	DATAERR
?SEP4:	RET

;
; Test for a comma
;
?COMMA:	LAHF
	PUSH	AX
	MOV	AL,BYTE PTR STRTYP
	CMP	AL,OTHER
	JNZ	?COMMA1

	MOV	AL,BYTE PTR STRING
	CMP	AL,','
?COMMA1:JZ	?COMMA2
	CALL	COMAERR			; comma expected

?COMMA2:POP	AX
	SAHF
	RET
;
; Test Rr
;
?RR:	MOV	CL,AL
	AND	AL,11111000B
	CMP	AL,00001000B
	MOV	AL,CL
	RET
;
; Test @Ri
;
?RI:	MOV	CL,AL
	AND	AL,11111110B
	MOV	AL,CL
	RET
;
; Test @Ri, Rr
;
?RRI:	CALL	?RR
	JNZ	?RRI1
	RET

?RRI1:	JMP	?RI
;
; Test Pp1
;
?PP1:	CMP	AL,19H
	JNZ	?PP11
	RET

?PP11:	CMP	AL,1AH
	RET
;
; Test Pp4
;
?PP4:	MOV	CL,AL
	AND	AL,11111100B
	CMP	AL,00011100B
	MOV	AL,CL
	RET
;
; Expect A,
;
ACOMMA:	CALL	OPRREG
	CMP	AL,7
	JZ	ACOMMA1
	CALL	REGERR

ACOMMA1:JMP	?COMMA
;
; Scan for a byte	AL=byte BX=addr CL=count, Z if found
;
SCAN:	PUSH	CX
	MOV	CH,AL
SCAN1:	CMP	AL,[BX]
	JZ	SCAN2

	INC	BX
	DEC	CL
	JNZ	SCAN1

	XOR	AL,AL
	INC	AL
SCAN2:	MOV	AL,CH
	POP	CX
	RET
;
; Evaluate word
;
EVALW:	CALL	EVAL
	MOV	BX,WORD PTR RESULT
	RET
;
; Evaluate byte
;
EVALB:	CALL	EVALW
	MOV	AL,BH
	OR	AL,AL
	JZ	EVALB1
	CALL	VALUERR

EVALB1:	MOV	AL,BL
	RET
;
; Get 16 bit operand BX
;
GETOPR:	PUSH	CX
	CALL	PARSE
	CALL	EVALW
	POP	CX
	RET
;
; Get 8 bit operand A
;
OPRBYT:	PUSH	CX
	CALL	PARSE
	CALL	EVALB
	POP	CX
	RET
;
; Get register operand A
;
OPRREG:	PUSH	CX
	CALL	PARSE
	CALL	?MNEM
	JZ	OPRREG1
	CALL	REGERR

OPRREG1:CMP	AL,16
	JZ	OPRREG2
	CALL	REGERR

OPRREG2:MOV	AL,CH
	LAHF
	PUSH	AX
	CALL	PARSE
	POP	AX
	SAHF
	POP	CX
	RET
;
; Reached end of operand
;
ENDOPR:	CALL	LABELL
	CALL	ALIGNPC
	JMP	ENDSTM
;
; Align pc to current value
;
ALIGNPC:MOV	BX,WORD PTR CURPC
	MOV	WORD PTR PC,BX
	RET
;
; Does current line have a label?
;
?LABEL:	MOV	BX,WORD PTR LABLPTR
	MOV	WORD PTR SYMPTR,BX
	JMP	?SYM
;
; If label present, assign PC (pass 1), or check its value (pass 2)
;
LABELL:	CALL	?LABEL			; label for this line?
	JNZ	LABEL1
	RET				; no, then exit

LABEL1:	MOV	BX,0			; clear label pointer
	MOV	WORD PTR LABLPTR,BX
	MOV	AL,BYTE PTR PASS	; which pass?
	OR	AL,AL
	JNZ	LABEL3

	CALL	GETNBL			; pass 1 - assign value of pc
	LAHF
	PUSH	AX
	AND	AL,00000111B		; label previously assigned?
	JZ	LABEL2
	CALL	LABLERR			; yes

LABEL2:	POP	AX
	SAHF
	OR	AL,00000001B		; set bit 0
	CALL	PUTNBL
	MOV	BX,WORD PTR PC		; and set value
	JMP	PUTSVAL

; pass 2 - check phase errors

LABEL3:	CALL	GETNBL
	AND	AL,00000111B		; encountered on previous pass?
	JNZ	LABEL4
	CALL	PHASERR			; no

LABEL4:	CALL	GETSVAL
	XCHG	BX,DX
	MOV	BX,WORD PTR PC		; value changed?
	CMP	DX,BX
	JZ	LABEL5
	CALL	PHASERR			; yes
LABEL5:	RET

;
; Send word BX in low/high format to hexfile and output line
;
OUTWRD:	PUSH	BX
	MOV	AL,BL
	CALL	OUTBYT
	POP	BX
	MOV	AL,BH
;
; Send byte A to hexfile and output line
;
OUTBYT:	MOV	CH,AL
OUTBYT1:MOV	AL,BYTE PTR PASS
	OR	AL,AL
	MOV	AL,CH
	JZ	OUTBYT3

	PUSH	CX
	CALL	WRTHEX			; send byte to hex file

	MOV	AL,BYTE PTR PCFLD	; if PC field blank then
	CMP	AL,' '			; new line, insert pc
	JNZ	OUTBYT2
	CALL	INSPC

OUTBYT2:MOV	AL,BYTE PTR HEXPOS	; enough room?
	CMP	AL,PCFLD+15-LINEBUF
	POP	CX
	JNC	OUTBYT3

	MOV	AL,CH			; insert hex byte
	CALL	INSBYT
OUTBYT3:MOV	BX,WORD PTR CURPC	; incr program counter
	INC	BX
	MOV	WORD PTR CURPC,BX
	RET
;
; Insert byte (A) into output line as ascii hex
;
INSBYT:	LAHF
	PUSH	AX			; get upper nibble
	RCR	AL,1
	RCR	AL,1
	RCR	AL,1
	RCR	AL,1
	AND	AL,00001111B
	CALL	INSBYT1			; to hex and line
	POP	AX
	SAHF
	AND	AL,00001111B		; get lower nibble
INSBYT1:CALL	BIN2HEX
	MOV	BX,OFFSET HEXPOS
	MOV	DL,[BX]
	MOV	DH,0
	INC	BYTE PTR [BX]
	MOV	BX,OFFSET LINEBUF
	ADD	BX,DX
	MOV	[BX],AL
	RET
;
; Covert binary A to ascii hex
;
BIN2HEX:ADD	AL,'0'
	CMP	AL,'9'+1
	JNC	BIN2HEX1
	RET

BIN2HEX1:
	ADD	AL,7
	RET
;
; Insert program counter into pc field of output line
;
INSPC:	MOV	BX,WORD PTR PC		; current pc
;
; Insert word BX as ascii into pc/equate field of output line
;
INSWRD:	XCHG	BX,DX
	MOV	BX,OFFSET HEXPOS	; save current 'hex' position
	PUSH	BX
	MOV	BYTE PTR [BX],PCFLD-LINEBUF ; set to 'pc' field
	MOV	AL,DH
	PUSH	DX
	CALL	INSBYT
	POP	DX
	MOV	AL,DL
	CALL	INSBYT
	POP	BX
	INC	BYTE PTR [BX]		; leave a blank
	RET
;
; Set error flags
;
SYNTERR:MOV	AL,'S'			; syntax error
	CALL	FLGERR
	JMP	SKIP			; don't return, skip to next statement

VALUERR:LAHF
	PUSH	AX			; value error
	MOV	AL,'V'
	JMP	ERROR

OVFLERR:LAHF
	PUSH	AX			; overflow error
	MOV	AL,'O'
	JMP	ERROR

EXPRERR:LAHF
	PUSH	AX			; expression error
	MOV	AL,'E'
	JMP	ERROR

REGERR:	LAHF
	PUSH	AX			; register error
	MOV	AL,'R'
	JMP	ERROR

DATAERR:LAHF
	PUSH	AX			; data error
	MOV	AL,'D'
	JMP	ERROR

PHASERR:LAHF
	PUSH	AX			; phase error
	MOV	AL,'P'
	JMP	ERROR

LABLERR:LAHF
	PUSH	AX			; label error
	MOV	AL,'L'
	JMP	ERROR

BALERR:	LAHF
	PUSH	AX			; balance error
	MOV	AL,'B'
	JMP	ERROR

UDEFERR:LAHF
	PUSH	AX			; undefined error
	MOV	AL,'U'
	JMP	ERROR

COMAERR:LAHF
	PUSH	AX			; comma error
	MOV	AL,'C'
	JMP	ERROR

NIMERR:	LAHF
	PUSH	AX			; not implemented error
	MOV	AL,'N'

ERROR:	CALL	FLGERR
	POP	AX
	SAHF
	RET

; flag error in current line

FLGERR:	PUSH	CX
	MOV	CH,AL
	CALL	?ERROR
	JNZ	FLGERR1			; error already flagged

	MOV	AL,CH
	MOV	BYTE PTR ERRFLD,AL
FLGERR1:POP	CX
	RET
;
; Test for error flag in current line
;
?ERROR:	MOV	AL,BYTE PTR ERRFLD
	CMP	AL,' '
	RET
;
; Subtract BX from DX, result in BX
;
DIFF:	MOV	AX,DX
	SUB	AX,BX
	MOV	BX,AX
	RET
;
; Block move  BX=source, DX=dest, CH=count
;
MOVE:	INC	CH
MOVE1:	DEC	CH
	JNZ	MOVE2
	RET

MOVE2:	MOV	AL,[BX]
	XCHG	BX,DX
	MOV	[BX],AL
	XCHG	BX,DX
	INC	BX
	INC	DX
	JMP	MOVE1
;
; Call MS-DOS function
;
DOS:	MOV	AL,0
	INT	21H
	RET
;
; Pre-initialised data area
;
SYMTOP	DW	TOP		; top of symbol table
SYMTBL	DW	TOP		; beginning of symbol table
PRNPTR	DW	0		;    "	     prn buffer
HEXPTR	DW	0		;    "	     hex buffer
ERRCNT	DW	0		; error count
PRC	DB	1		; program return code
;
; File FCBs
;
SRCFCB	DB	9 DUP (0)	; fcb for source file
	DB	'A48'
	DB	25 DUP (0)

PRNFCB	DB	9 DUP (0)	; fcb for prn file
	DB	'PRN'
	DB	25 DUP (0)

HEXFCB	DB	9 DUP (0)	; fcb for hex file
	DB	'HEX'
	DB	25 DUP (0)

LINNO	DB	'    0'		; string holding line number
;
; Messages
;
MSG1	DB	'MS-DOS 8048/41 Assembler ','0'+REL,'.','0'+REV,0
MSG1A	DB	'  '
	DATE
	DB	CR,LF,CR,LF,0
MSG2	DB	'No source file present',CR,LF,0
MSG3	DB	'No directory space',CR,LF,0
MSG4	DB	'Source file name error',CR,LF,0
MSG5	DB	'Source file read error',CR,LF,0
MSG6	DB	'Output file write error',CR,LF,0
MSG7	DB	'Cannot close files',CR,LF,0
MSG8	DB	'End of assembly',CR,LF,0
MSG9	DB	'Symbol table overflow',CR,LF,0
MSG11	DB	'Errors    0',CR,LF,0
MSG11A	EQU	$-4
MSG12	DB	'   Page 000   ',0 	; header
MSG12A	EQU	$-5
MSG12B	EQU	$-1
	DB	64 DUP (?)		; space for title
;
; Un-initialised data area
;
DOSVER	DB	1 DUP (?)	; DOS version
MEMTOP	DB	2 DUP (?)	; top of memory
DEF	DB	1 DUP (?)	; default drive
SRC	DB	1 DUP (?)	; source file drive
PRN	DB	1 DUP (?)	; print file drive
HEX	DB	1 DUP (?)	; hex file drive
PASS	DB	1 DUP (?)	; assembler pass
PC	DB	2 DUP (?)	; program counter for instruction
CURPC	DB	2 DUP (?)	; current location counter
EXEC	DB	2 DUP (?)	; execution address
PRVCHAR	DB	1 DUP (?)	; previous character from source file
CHAR	DB	1 DUP (?)	; current character from source file
LINES	DB	1 DUP (?)	; count of lines output to prn file
STRTYP	DB	1 DUP (?)	; type of string
STRVAL	DB	2 DUP (?)	; string evaluated to 16 bits
RESULT	DB	2 DUP (?)	; result of eval routine
PFLAG	DB	1 DUP (?)	; parse flag
SYMPTR	DB	2 DUP (?)	; symbol entry
HEXADDR	DB	2 DUP (?)	; current hex address pointer
NUMHEX	DB	1 DUP (?)	; number of hex chars in buffer
HEXREC	DB	1 DUP (?)	; hex record type, 0=data 1=end
HEXPOS	DB	1 DUP (?)	; position of next hexbyte in prn file
SRCPTR	DB	2 DUP (?)	; pointer to source buffer
FNDEPTH	DB	1 DUP (?)	; depth of function/precedence stacks
PRMDPTH	DB	1 DUP (?)	; depth of parameter stack
DIVTMP	DB	2 DUP (?)	; division temp save
DIVCNT	DB	1 DUP (?)	; division shift count
LABLPTR	DB	2 DUP (?)	; pointer to current label
HASHVAL	DB	1 DUP (?)	; symbol name hashed to value 0-127
NUMBASE	DB	1 DUP (?)	; current number base
HEXLIST	DB	16 DUP (?)	; buffer for 16 hex values
FUNCSTK	DB	10 DUP (?)	; function stack
PRECSTK	DB	10 DUP (?)	; precedence stack
PRMSTK	DB	16 DUP (?)	; parameter stack (8 max)
IFMODE	DB	1 DUP (?)	; IF mode
IFCOUNT	DB	1 DUP (?)	; IF count
NESTLVL	DB	1 DUP (?)	; nested IF level
NESTSTK	DB	8 DUP (?)	; nested IF stack
	DB	64 DUP (?)	; stack space
STACKK	DB	2 DUP (?)	; old stack
LINELEN	DB	1 DUP (?)	; current position in line buffer
LINEBUF	DB	PWIDTH DUP (?)	; line buffer (must follow length)

NUMFLD	EQU	LINEBUF		; line number field
ERRFLD	EQU	LINEBUF+6	; error field
PCFLD	EQU	LINEBUF+8	; pc/equate field
SRCFLD	EQU	LINEBUF+24	; source field

STRLEN	DB	1 DUP (?)	; string length
STRING	DB	64 DUP (?)	; string buffer (must follow length)
SRCBUF	DB	NASM*128 DUP (?) ; source buffer space
PRNBUF	DB	NPRN*128 DUP (?) ; prn buffer space
HEXBUF	DB	NHEX*128 DUP (?) ; hex buffer space
LNKTBL	DB	256 DUP (?)	; symbol link table

TOP	EQU	$		; beginning of symbol table

MAIN	ENDS

	END	START

